#!/usr/bin/perl -w

sub usage(){
    print "Usage: $0 -skd <skd file> [-st <id>] [-dir <a[:b:c:d]>] [-fn <n>]\n";
    print " \t[-host <host>:<port>] [-n <nch>] [-mode <Msps>:<bit>:<ch>]\n";
    print " \t[-sim ]  [-exe] \n";
    print " Auto Observation support tool for recording with ADS2000+PC-VSI\n";
    print " -st: Station ID used in the sched file , Default is 'O' \n";
    print " -dir: Output data dirs, if less than four are specified, all the data is stored on the same one directory\n";
    print " -host: Data receiver host(default=localhost):<port(default=9903)> \n";
#    print " -n: Total Number of channel (1/4/8/16). default=4. \n";
#    print " -mode: Recording mode. default=16:1:4 \n";
    print " -fn: File naming convention: (default 2)\n";
    print "     1 : XDDDNNNN.vdif (default)\n";
    print "    -1 : XDDDNNNN.#ch.vdif \n";
    print "     2 : sidDDDHHMMSSG.vdif \n";
    print "    -2 : sidDDDHHMMSSG.#ch.dat\n";
    print "     3 : expid_sidG_scanid_YYYYDDDHHMMSS.k5 \n";
    print " -sim: Simulation mode\n";
    my ($Date, $Revision, ) = ("Date", "Revision");
    $, = '';
    print "$Revision: 1.9 $, $Date: 2011/09/09 04:12:33 $, \n";
    print " -----------------------------------------\n";
    exit(1);
}

use lib "$ENV{HOME}/prog/perl/module";
use Sched_db;
use Time_lib;
use IO::Socket;  
use Term::ANSIScreen qw/:color :cursor :screen :keyboard/;
use Time::HiRes qw(clock_gettime clock_getres CLOCK_REALTIME usleep);

# k5vsi-vssp-dim -m '08Msps/2bit/04ch' -o test-vssp.dat1:test-vssp.dat2:test-vssp.dat3:test-vssp.dat4 -g 0:1:2:3 -t 10s -d

my $cnt=0;
my $extra_end = 7;   ## extra delay sec for VSI
{
    usage() if(@ARGV < 1); 
    my ($skdfile, $stid);
    my @odir = qw(. . . .);
    my $exflag =0;
    my $fn=2;
    my $sim = 0;
    my $Msps=2048; # Msps
    my $bit = 1; # bit
    my $gch = 1; # $channel in a group
    my $nch= 1;   # Total number of channel
    my $host="localhost";
    my $port=9903;

    while(@ARGV){
	my $opt = shift;
	if($opt =~/-skd/){
	    unless(@ARGV){ usage();}
	    $skdfile = shift;
	}elsif($opt =~/-st/){
	    unless(@ARGV){ usage();}
	    $stid = shift;
	}elsif($opt =~/-dir/){
	    unless(@ARGV){ usage();}
	    my $odir = shift;
	    my @terms = split(/:/,$odir);
	    if( @terms < 4){ 
		@odir = ($terms[0],$terms[0],$terms[0],$terms[0]);
	    }else{@odir = @terms; }
	}elsif($opt =~/-fn/){
	    unless(@ARGV){ usage();}
	    $fn = shift;
	}elsif($opt =~/-host/){
	    unless(@ARGV){ usage();}
	    my @terms = split(/:/,shift);
	    if(@terms<2){
		$host = $terms[0];
	    }else{
		$host = $terms[0];
		$port = $terms[1];
	    }
	}elsif($opt =~/-mode/){
	    unless(@ARGV){ usage();}
	    my $tmp = shift;
	    my @terms =  split(/:/,$tmp);
	    if(@terms < 3){
		print "E Mode format is wrong: $tmp \n";
		usage();
	    }else{
		$Msps = $terms[0];
		$bit  = $terms[1];
		$gch  = $terms[2];
	    }
	}elsif($opt =~/-n/){
	    unless(@ARGV){ usage();}
	    $nch = shift;
	}elsif($opt =~/-sim/){
	    $sim = 1;
	}elsif($opt =~/-exe/){
	    $exflag=1;
	}
    }

    unless( defined($skdfile)){
	print " E Missing sched file.!!\n";
	usage();
    }
    #unless($nch % $gch == 0){
	#print "E Total Number of Channel is not multiple of #group channels.\n";
	#usage();
    #}
    
    my $skd = new Sched_db;
    $skd->Read($skdfile);
    my $nskd = $skd->Nskd();
    my $srate = $skd->Srate();
    my $expcode = $skd->Expr();

    my $ngrp = $nch/$gch;

    #&communicate($host, $port,"data_out=multi:$ngrp;");
    #for(my $i =0;  $i < $ngrp; $i++){
	#my $mode = sprintf "%d:%d:%d:%d:%d", $i, $Msps, $bit, $gch, $i*$gch;
	#&communicate($host, $port,"modegrp=$mode;");
    #}
	my $mode = sprintf "%d:%d:%d", $Msps, $bit, $nch;
	&communicate($host, $port,"mode=$mode;");

    my $opt0 = sprintf("-m '%02dMsps/2bit/%02dch' -g 0:1:2:3",
		       $srate, $nch);
    my $catch_flag=0;
    my $tofst = 0;
    $tnow = time();
    if($sim ){ 
	$t1 = $skd->Skd_start(0,3)->UXtime();
	$tofst = $tnow - $t1;
    }

    for(my $iscan=0; $iscan < $nskd; $iscan++){
	my $srcname = $skd->Skd_src($iscan);
	my $dur = $skd->Skd_dur2($iscan,$stid);
	next if( $dur < 0);
	$t1 = $skd->Skd_start($iscan,3);
	$t2 = $skd->Skd_stop2($iscan,$stid,3);
	$t1 = $t1 + $tofst;
	$t2 = $t2 + $tofst;
	$tnow = clock_gettime(CLOCK_REALTIME);
	my $twait = $t1->UXtime() - $tnow;
	my $twait2 = $t2->UXtime() - $tnow;

	my $tm = new Time_lib;
	if( $twait > 1){
	    print "\n\n\n" if ($catch_flag==0);
	    $catch_flag=1;
	    while($twait > 0.5){
		print up(1);
		my $tw = $twait > 1 ? 1.e6 : ($twait-0.2)*1.e6;
		$tm->Set_time_now();
		printf "Now(%s) waiting (%d sec) for Start:%s\n",
			$tm->Str_fmt3(2), $twait,$t1->Str_fmt3(2);
		usleep($tw);
		$tnow   = clock_gettime(CLOCK_REALTIME);		
		$twait  = $t1->UXtime() - $tnow;
		$twait2 = $t2->UXtime() - $tnow;
	    }
	}
	if($twait2 < 0){
	    print "$iscan ";
	    next;
	}
	my @fname;
	for(my $igrp=0; $igrp<$ngrp; $igrp++){
	    $fname[$igrp] = $odir[$igrp] . "/"
		. &Naming($skd,$iscan,$stid,$igrp,$fn);
	    my $outspec = sprintf("data_out=FILE%d:%s;\n",
		$igrp,$fname[$igrp]);
	    print "$outspec \n";
	    &communicate($host, $port,$outspec);
	}

	if($exflag){
	    usleep(200000);
	    &communicate($host, $port,"in2net=on;");
	    while($twait2 >0){
		my $tw = $twait2 > 1 ? 1.e6 : ($twait2)*1.e6;
		$tm->Set_time_now();
		print up(1);
		printf "Now(%s) waiting (%d sec) for End:%s   \n",
		$tm->Str_fmt3(2), $twait2,$t2->Str_fmt3(2);
		usleep($tw);
		$tnow   = clock_gettime(CLOCK_REALTIME);		
		$twait2 = $t2->UXtime() - $tnow + $extra_end;
	    }
	    &communicate($host,$port, "in2net=off;");
	}else{
	    print "***** This is testing mode *************\n";
	    print "****** use -exe  for real execution. ***\n"
	}
    }
    
}
    


sub Naming($$$;$){
    my ($skd, $i, $id, $igrp,$fn) = @_;
    $fn = 1 unless(defined($fn));
    my $t1 = $skd->Skd_start($i,3);
    my $doy = $t1->Doy();
    my @grp = qw(a b c d);
    my @ydhms = $t1->YDHms();
    my $fname;    
    if($fn == 1){
	$fname = sprintf("%s%03d%04d.vdif",$id,$doy,$i);
    }elsif($fn == -1){
	$fname = sprintf("%s%03d%04d.%d.vdif",$id,$doy,$igrp,$i);
    }elsif($fn == 2){
	$fname = sprintf("%s%03d%02d%02d%02d%s.vdif",
			 $id,$ydhms[1],$ydhms[2],$ydhms[3],$ydhms[4],
			 $grp[$igrp]);
    }elsif($fn == -2){
	$fname = sprintf("%s%03d%02d%02d%02d%s.%d.dat",
			 $id,$ydhms[1],$ydhms[2],$ydhms[3],$ydhms[4],
			 $grp[$igrp], $igrp);
    }elsif($fn == 3){
	my $expr = $skd->Expr();
	my $src =  $skd->Skd_src($i);
	$fname = sprintf("%s_%s%s_%s_%04d%03d%02d%02d%02d.k5",
			 $expr, $id, $grp[$igrp],$src,$ydhms[0],
			 $ydhms[1],$ydhms[2],$ydhms[3],$ydhms[4]);
    }
    return $fname;
}


sub communicate($$$){
    my ($host,$port,$line) = @_;

    my $remote = IO::Socket::INET->new(Proto    => "tcp",
				       PeerAddr => $host,
				       PeerPort => $port );
    if( ! $remote){	die "Could not connect to $host:$port \n";    }
    select($remote); $|=1; # stop buffering 
    select(STDOUT); $|=1; # stop buffering


    print STDERR "Sent: $line\n";
    last unless(defined($remote));
    print $remote "$line\n";
    last unless(defined($remote));
    
    my $res = <$remote>;
    next unless(defined($res));
    chomp($res);
    print STDERR "Res[$cnt]: $res\n" if(defined($res));
    $cnt++;
    close($remote);
    
}
