#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>

#include "VDIF_Frame.hpp"
/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"

const char* version = "$Revision: 273 $ -Latest Update: $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $";

void usage(){
  cerr << "Usage: " <<   me << " -f <filename>  [-i <byte>] [-n <int>] [-t <1|2|3|4|5|6>] [-d <int>]\n" ;
  cerr << "             [-w <fname:threadid>] \n";
  cerr << "-f : file name \n";
  cerr << "-i : Forced Interval of frame [byte].\n";
  cerr << "-t : header print type.[default=1]\n";
  cerr << "-n : Number of frames to show. [default=10]\n";
  cerr << "-d : Thread Number. [default=0]\n";
  cerr << "-w : Write out Specified Thread ID data stream to a file. \n";
  cerr << "  Ver.:" << version << endl;
  exit(1);
}

class VDIF_FixRead{
  FILE *fp_;
  VDIF_Hdr::Prnt_t hdp_type;
  static VDIF_Hdr::Prnt_t hdpt_arr[6];
  long intvl_byte;
  int nframe_disp;
public:
  VDIF_FixRead():fp_(0),hdp_type(VDIF_Hdr::CONST),intvl_byte(0),nframe_disp(10){};
  int Open(const char* fname){
    fp_=fopen( fname, "r");
    if(fp_== NULL){
      printf("Error, Failed to open file:%s \n", fname );
      return -1;
    }
    else return 0;
  };
  void SetHdrDispType(const int a);
  void SetNframeDisp(const int a){ nframe_disp=a;}
  void SetFrameIntvl(const long a) {intvl_byte=a;}
  int Monitor(const char *fname, const int thid=0);
  int Write1(const char* fname, const char* wfname,const int thid);
};

VDIF_Hdr::Prnt_t VDIF_FixRead::hdpt_arr[]={VDIF_Hdr::ALL, VDIF_Hdr::CONST,VDIF_Hdr::VAR,VDIF_Hdr::ONELINE,VDIF_Hdr::ALL_ONE,VDIF_Hdr::RAW,};
inline void VDIF_FixRead::SetHdrDispType(const int a){ hdp_type = hdpt_arr[a]; };


int VDIF_FixRead::Monitor(const char *fname, const int thid){
  VDIF_Hdr *phdr=0;
  const int bsize=64;
  unsigned char* pbuf = new unsigned char[bsize];
  phdr= (VDIF_Hdr*) pbuf;
  
  if(Open(fname) < 0) return -1;

  fread(pbuf,bsize,1,fp_);
  phdr->Print_hd(stdout,hdp_type);
  
  if( intvl_byte == 0) intvl_byte= phdr->Framesize();
  long jump = intvl_byte - bsize;  

  const int ntry=100000;
  for(int i=0; i< nframe_disp; i++){
    int itry=ntry;
    while(itry >0){
	fseek(fp_,jump,SEEK_CUR);
	fread(pbuf,bsize,1,fp_);
	if(phdr->ThreadID() != thid) itry--;
	else break;
    }
    if( itry < 1){
      fprintf(stderr,"E Thread(%d) is not found within %d frames.\n",
	      thid, ntry);
      break;
    }
    phdr->Print_hd(stdout,hdp_type);
  }
  
  fclose(fp_);
  return 0;
}

int  VDIF_FixRead::Write1(const char* fname, const char* wfname, const int thid){

  FILE* wfp = fopen(wfname,"w");
  if(wfp==NULL  ){ return -1;}

  double wbyte=0.0;
  VDIF_Hdr *phdr=0;
  if( intvl_byte == 0) intvl_byte = phdr->Headsize();
  unsigned char* pbuf = new unsigned char[intvl_byte];
  phdr= (VDIF_Hdr*) pbuf;
  
  if(Open(fname) < 0) return -1;
  fread(pbuf,intvl_byte,1,fp_);
  if( intvl_byte == phdr->Headsize() ) intvl_byte= phdr->Framesize();
  rewind(fp_);
  double sod=-1.0;
  fprintf(stderr," Start Writing %s [ThreadID=%d] from %s ============\n", 
	  fname, thid,  phdr->Epoch().Print_time(14).c_str()) ; 
  while( fread(pbuf,intvl_byte,1,fp_) > 0 ){
    if(phdr->ThreadID() == thid){
      //      phdr->Set_FrameLen(intvl_byte);
      fwrite(pbuf,intvl_byte,1,wfp);
      wbyte += 1.e-6* intvl_byte;
      if(sod != phdr->Epoch().Sod()){
	fprintf(stderr," %s %.2lf MB write. \r", 
		phdr->Epoch().Print_time(14).c_str(), wbyte); 
      }
    }
  }
  fprintf(stderr,"============ END ============================\n");
  fclose(fp_);
  fclose(wfp);
  if(wbyte ==0.0){
    return -1;
  }else{
    return 0;
  }
}

int main(int argc, char *argv[]){
  SetMyname(argv[0]);

  VDIF_FixRead fix;
  
  string fname, wfname;
  int thid=0;
  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:i:n:t:w:d:")) { 
    case 'h':  usage(); break;
    case 'f':    fname = optarg;      break;
    case 'd':    thid = strtol(optarg,0,10);      break;
    case 'i':   fix.SetFrameIntvl( strtol(optarg,0,10) );     break;
    case 't': {int tmpi = strtol(optarg,0,10) % 6 ;     fix.SetHdrDispType(tmpi);  break;}
    case 'n':   fix.SetNframeDisp( strtol(optarg,0,10)) ;     break;
    case 'w':   wfname=optarg;     break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr,  "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number in command line \n", me); 
      usage();
      break; 
    }
  }
  
  if( fname.empty() ) usage();
  
  if( wfname.empty()){
    fix.Monitor( fname.c_str(), thid );
  }else{
    if( fix.Write1(fname.c_str(), wfname.c_str(),thid) < 0){
      fprintf(stderr,"Err: Write File(%s) Failed.\n", wfname.c_str());
      exit (1);
    };
  }

    
}
