#include <cstdio>
#include <cstdlib>
#include <iostream>
using namespace std;

#include "VDIF_Frame.hpp"

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"
const char* version = "$Revision: 292 $ -Latest Update: $Date: 2018-04-21 12:00:25 +0900 (土, 21  4月 2018) $";

void Print_dat( uint16_t *ptr, const int i);

void usage(){
  cerr << "Usage: " <<  me << " -f <filename> [-m <0|1|2|3>] [-n <num>] \n";
  cerr << "   [-d <lines>] [-E] [-s <sec>] [-e <yyyy.ddd.HH:MM:SS> ] [-v < -2 - 2 >]\n";
  cerr << "  Reading Elecs VDIF file \n";
  cerr << "  -E: Read Option for Elecs VDIF file. \n";
  cerr << "  -e: Specify the epoch to display. \n";
  cerr << "  -m: Display mode: Defalut 2\n";
  cerr << "    0: Print each frame(o/.) \n";
  cerr << "    1:Print Every Headr data. \n";
  cerr << "    2:Print Headr data Every second. \n";
  cerr << "    3:Print Headr data begning and end of each sec. \n";
  cerr << "  -d: Number of data lines(16 Byte) to print. default=3.\n";
  cerr << "  -n: Number of lines to print. default=10\n";
  cerr << "  -s: Skip first <sec> of data before display. \n";
  cerr << "  -v: " << Verb_str() << endl;
  cerr << "  Ver.:" << version << endl;
  exit(1);
}

int main(int argc, char *argv[]){
  SetMyname(argv[0]);
  Time_dat td;
  string fname, tepoch;
  int mode=2;
  int k=0;
  int nline=10;
  int print_data=3, skip_sec=0;
  bool elecs_flag=false;
  
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:v:m:n:d:Es:e:")) { 
    case 'h':  usage(); break;
    case 'f':    fname = optarg;      break;
    case 'E':    elecs_flag=true;      break;
    case 'e':    tepoch=optarg; td.Set_time(tepoch,5,":.");     break;
    case 'n':    nline = strtol(optarg,0,10);      break;
    case 's':    skip_sec = strtol(optarg,0,10);      break;
    case 'd':    print_data = strtol(optarg,0,10);      break;
    case 'v':    verbose_no=  (Verbose_level_t)strtol(optarg,0,10);      break;
    case 'm':    mode = strtol(optarg,0,10);      break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(fname.empty()) usage();

  VDIF_Frame  vdif;
  
  if( ! vdif.ROpen(fname)){
    cerr << "Error in opening file: " << fname << endl;
    exit(1);
  }
 
  if( !vdif.find_hdr() ){
    cerr << "E Failed to find VDIF Header.\n";
    exit(1);
  }else{
    fprintf(stderr,"Hdr-Found:Offset=%d \n",vdif.Hdr_Offset() );
  }

  long framesize = vdif.Framesize();
  long datsize = framesize - vdif.Headsize();
  //  long nframe_sec = vdif.Nframe_sec_estimate();
  long nframe_sec = vdif.Nframe_sec();
  int nch = vdif.Hdr().Nch();
  int qbit = vdif.Hdr().Qbit();
  fprintf(stderr,"----------------------------------------\n");
  fprintf(stderr,"framesize: %5ld [byte] datasize: %5ld [byte]\n", framesize, datsize);
  fprintf(stderr,"Nch      : %4d        Qbit     : %2d  Estimated Frame/sec : %6ld\n", nch, qbit,nframe_sec );
  fprintf(stderr,"----------------------------------------\n");
  uint8_t frame_buf[framesize+128];
  VDIF_Hdr * phdr = (VDIF_Hdr *)frame_buf;
  
  vector <uint8_t> rbuf;
  ReturnFrameInfo retinfo;
  FILE *fp=stdout;

  vdif.Rewind();
  long cnt=0;
  int lcnt=0;
  long tcnt=0;
  float frame_ratio=0;
  uint16_t *pdat=0, *pdat0= (uint16_t *)(frame_buf+vdif.Headsize());
  print_data = print_data > datsize/16 ? datsize/16 : print_data;
  Time_dat t1;
  //  Time_dat t2=t1;
  bool first_flag=true;
  VDIF_Hdr prev_hdr;

  pdat = (uint16_t *)(frame_buf+vdif.Headsize());
  while(! vdif.EndofFile_Yes() ){
    if(elecs_flag && tcnt == ELECS_NO ) {
      vdif.Read_Bytes(frame_buf, framesize+128);
      tcnt=0;
    }else
      vdif.Read_Bytes(frame_buf, framesize);
    cnt++;
    tcnt++;    
    if(! tepoch.empty() && phdr->HD_Match()){ 
      if( td > phdr->Epoch() ){
	if( phdr->iFrame() % 1000 ==0)
	  fprintf(stderr,"Now: %s waiting for %s \r", 
		  phdr->Epoch().Print_time(15).c_str(),
		  td.Print_time(15).c_str());
	continue;
      }
    }
    pdat = pdat0;
    switch(mode){
    case 0: // Every frame with mark.
      if(phdr->HD_Match()){ putc('o',fp);}else{ putc('.',fp); }
      //      cnt++;
      if(cnt%100 ==0){ 
	lcnt++;
	putc('\n',fp);
	if(nline > 0 && nline < lcnt) exit(1);
      }
      break;
    case 1: // Every Frame 
      if(phdr->HD_Match()){ cout << phdr->Str_hd(VDIF_Hdr::ONELINE) << endl;}
      
      for( int i=0;i< print_data; i++) {Print_dat(pdat,i); pdat += 8;}
      //      cnt++;
      if(nline > 0 && nline < cnt) exit(1);
      break;
    case 2:  // Every second
      //      cnt++;
      if( phdr->HD_Match() ){ 
	if (t1 != phdr->Epoch() || first_flag){
	  
	  if( ! first_flag){
	    frame_ratio=(float)cnt/nframe_sec * 100.0;
	    fprintf(stdout,"%s %6ld %6.2f %% \n", 
		    prev_hdr.Str_hd(VDIF_Hdr::ONELINE).c_str(), cnt, frame_ratio);
	  }
	  fprintf(stdout,"%s \n",  phdr->Str_hd(VDIF_Hdr::ONELINE).c_str());
	  for( int i=0; i<print_data; i++){ Print_dat(pdat,i); pdat+=8;}

	  t1 = phdr->Epoch();
	  first_flag=false;
	  cnt=0;
	}
      }
      break;
    case 3: // Every 20 frame at first and end of every second.
      { 
	bool pflag=false;
	if(phdr->iFrame() < nline ||  phdr->iFrame() > nframe_sec -nline ) pflag=true;
	if(phdr->HD_Match()  &&  (t1 != phdr->Epoch() )){
	  frame_ratio=(float)cnt/nframe_sec * 100.0;
	  t1 = phdr->Epoch();
	  cnt=0;
	  cout << "------------------------------------------------------------" << endl;
	  lcnt++;
	  //	  if(nline > 0 && nline < lcnt) exit(1);

	}
	if(pflag){
	  fprintf(stdout,"%s %6ld \n",
		  phdr->Str_hd(VDIF_Hdr::ONELINE).c_str(), cnt);
	  
	  for( int i=0; i<print_data; i++){ Print_dat(pdat,i); pdat+=8;}
	}
      }
      //      cnt++;
      break;

    } // Switch 
    
    prev_hdr = *phdr;
  } // while 


}

void Print_dat(uint16_t *ptr, const int iser){
  fprintf(stdout,"%4d ", iser);
  for(int i=0; i<8; i++, ptr++)
    fprintf(stdout,"  %04X",(unsigned int)*(ptr) );
  //    fprintf(stderr,"  %02X%02X",(unsigned int)*(ptr),(unsigned int)*(ptr+1) );
  putc('\n',stdout);
}
