#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "time_lib.hpp"

#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "conf_utl.hpp"
extern "C"{
#include <unistd.h>
#include <sys/time.h>
}

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"
const char* version = "$Revision: 273 $ -Last Update: $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $";


Config_set vdif_conf;

void LogSetup(const int verbose_level);

void usage(){
  string version_str = version;
  cerr << "Usage: " << me << " -f <filename> [-o <outdir>] \n";
  cerr << "   [-v <level>] [-s <stanme>:<ID>] [-n <sec>]\n";
  cerr << "   [-c <first ch>]\n";
  cerr << " Extracting 4 channel of data and output VDIF file. Data must be 2bit sample data.\n";
  cerr << "-f : input VDIF file.\n";
  //  cerr << "-v : Verbose level -1,..,4.\n";
  cerr << "-o : output directory of VDIF file. \n";
  cerr << "-y : Specify year for  detection of VDIF header.\n";
  cerr << "-s : Station Name and ID.\n";
  cerr << "-c : Channel to be extected are <n>,<n+1>...<n+3>. n=0,1,2...\n";
  cerr << "     channel must be multile of 4.\n";
  cerr << "   Version: " << version_str << endl;
  //  cerr << "-z : Size of data reading at once.(Default= 3 MByte.)\n";
  exit(1);
}

int main(int argc, char *argv[]){
  SetMyname(argv[0]);
  string stdummy;
  long nsec = 0;
  string infname, outfile;
  int year=0;
  int firstch=0;

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:o:n:y:s:c:")) { 
    case 'h':  usage(); break;
    case 'f':    infname = optarg;      break;
    case 'y':    year = strtol(optarg,0,10); break;
    case 'c':    firstch = strtol(optarg,0,10); break;
    case 'n':    nsec = strtol(optarg,0,10); break;
    case 's':    stdummy = optarg;       break;
      //   case 'v':    verbose_level = strtol(optarg,0,10);      break;
    case 'o':    outfile = optarg;     break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  char stname[30]="", stid[5]="";
  if( ! stdummy.empty()){
    char buf[80];
    strcpy(buf,stdummy.c_str());
    char *cp = strchr( buf, ':');
    if(cp != NULL){
      *cp='\0';
      strncpy(stname,buf, 30);
      cp++;
      strncpy(stid,cp,2); stid[2]='\0';
    }
  }
  
  if(infname.empty() || outfile.empty()) usage();
  //  LogSetup(verbose_level);

  struct timeval tv0, tv1;   
  double total_wsize=0;
  int rd_nframe=10;
 
  try{
    gettimeofday(&tv0,0); // Processing start time
    
    VDIF_Frame  vdif;
    if( ! vdif.ROpen(infname)){
      string msg =  "Error in opening file: " +  infname;
      throw msg;
    }
    
    vector <VDIF_Hdr> hdrarr(rd_nframe);
    long nframes = vdif.Read_1sec_hdr( hdrarr,rd_nframe);
    if(nframes < 1){
      string msg = "E Failed to find VDIF Header.";
      throw msg;
    }
    
    long framesize = vdif.Framesize();
    long datsize = framesize - vdif.Headsize();
    int nch = vdif.Hdr().Nch();
    int qbit = vdif.Hdr().Qbit();
    //    long nframe_sec = vdif.Nframe_sec_estimate();
    long nframe_sec = vdif.Nframe_sec();
    long srate = vdif.Hdr().Srate();
    double drate = (double)srate/1000000*nch*qbit/8; // MB/sec    
    if(qbit < 2) {
      cerr << "--I am Expecting 2bit data. ---\n";
      usage();
    }
    if(vdif.Hdr().EDV() != EDV_NICT){
      drate = (double)nframe_sec*datsize;
      srate = 8*datsize*nframe_sec/nch/qbit;
    }
      
    fprintf(stdout,"Original VDIF: %s \n", infname.c_str()); 
    fprintf(stdout,"Output   VDIF: %s \n", outfile.c_str()); 
    fprintf(stdout,"Nch=%d Qbit=%d, Srate=%ld Msps.\n", nch,qbit,srate/1000000);
    fprintf(stdout,"DataRate:%.3lf MB/sec. Extracting %d-%d ch\n", 
	    drate,firstch, firstch+3);
    fprintf(stdout,"Nframe/sec [real/ideal]=%ld/%ld, framesize=%ld (%ld)Byte, Initial iframe=%ld \n",
	    nframes,   (long unsigned int)vdif.Nframe_sec(), 
	    framesize, datsize,  (long int)hdrarr[0].iFrame());
    long out_datasize=datsize/nch*4;
    long out_framesize = out_datasize + vdif.Headsize();

    for(int i=0; i<rd_nframe; i++){
      cout << hdrarr[i].Str_hd(VDIF_Hdr::ONELINE) << endl;
    }
    
    vdif.Rewind(VDIF_Frame::HEADER);

    unsigned char out_buf[out_framesize];
    unsigned char in_buf[framesize];
    VDIF_Hdr *phdr=(VDIF_Hdr*)in_buf;
    VDIF_Hdr *pohdr=(VDIF_Hdr*)out_buf;

    *pohdr = hdrarr[0];    
    VDIF_EUD_NICT eud_nict = hdrarr[0].EDV();
    eud_nict.Set_stname(stname);
    pohdr->exud_NICT=eud_nict;
    pohdr->Set_FrameLen(out_framesize);
    if(strlen(stid) >0) pohdr->Set_Stnid(stid);
    pohdr->Set_Sparam(4, qbit, srate);

    FILE* ofp = fopen(outfile.c_str(), "wb");
    if(ofp == NULL){
      cerr << "E Failed to open outfile(" << outfile << ")\n";
      exit(1);
    }
    
    Time_dat t0=hdrarr[0].Epoch(), t1=t0, t2=t0;
    long isec=-1; 
    long frm_cnt=0;
    int cpysize=qbit*4/8;
    int skipsize = nch*qbit/8;
    int ofst_byte=firstch*qbit/8;

    while(!vdif.EndofFile_Yes() && isec < nsec){
      if( vdif.Read_Bytes( in_buf, framesize ) == 0 ){
	if(vdif.EndofFile_Yes())
	  cerr << " End of File." << endl;
	else
	  cerr << "E : Failed to read." << endl;
	break;
      }
      if( (year >0 && phdr->HD_Match("Y", &year) )
	  || (year == 0 && phdr->HD_Match())){
	t2=phdr->Epoch();
	*pohdr = *phdr;
	pohdr->exud_NICT=eud_nict;
	pohdr->Set_FrameLen(out_framesize);
	if(strlen(stid) >0) pohdr->Set_Stnid(stid);
	pohdr->Set_Sparam(4, qbit, srate);
      
	uint8_t *rptr=in_buf +vdif.Headsize() + ofst_byte;
	uint8_t *wptr=out_buf +vdif.Headsize();
	for(long ds1=out_datasize; ds1 > 0; ds1-=cpysize){
	  memcpy(wptr,rptr,cpysize);
	  wptr+=cpysize;	rptr+=skipsize;
	}
	if(fwrite( (void*)out_buf, out_framesize, 1, ofp) < 1){
	  fprintf(stderr,"E Failed to write frame to outfile\n");
	  exit (1);
	}
	total_wsize +=out_framesize;
	frm_cnt++;
	if(nsec > 0) isec= t2-t0;
	if( t2 > t1 ){ t1=t2;
	  fprintf(stderr,"\r%s",t1.Print_time(14).c_str());
	}else  if(frm_cnt % 1000000 ==0) putc('.',stderr);
      
      }else{ // if HD not mached 
	if( ! vdif.find_hdr() ){
	  string msg="E Faile to find Header.";
	throw msg;
	}
      }
    } // end while
    fclose(ofp);

    gettimeofday(&tv1,0); // Processing End time
    double dt = (tv1.tv_sec - tv0.tv_sec)+  (tv1.tv_usec - tv0.tv_usec)*1.e-6 ;
    double wrate = total_wsize*1.e-6/dt; // MB/sec
    fprintf(stderr,"wrote  %.1lf MBytes. %ld frames\n", 
	    total_wsize*1.e-6, frm_cnt); 
    fprintf(stderr," %.3lf sec elapsed(%.1lf MB/s). Finished.\n", dt, wrate);
    cerr << "==========================================\n";
    
  }// End of Try
  
  
  catch(string msg){
    cerr << msg << endl;
  }
  
}
