/** 
 * @Author M.Sekido
 * @Version $Rev: 273 $
 * @Date  $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $
 * $Id: vdif_recv.cpp 56 2013-07-03 11:40:14Z vlbi $ 
 */
#include <iostream>
#include <string>
#include <vector>
using namespace std;
#include <boost/algorithm/string.hpp>

#include <time.h>
//#include <sys/types.h>
//#include <unistd.h>
#include <cstdio>
#include <cstring>
#include <cstdlib>

#include "VDIF_Frame.hpp"
///#include "VTProto.hpp"
#include "VDIFProto.hpp"
#include "Mk5B_Frame.hpp"
#include "Trans_ctl.hpp"
#include "Raw_Proto.hpp"
#include "InNet_mdl.hpp"
#include "Dumper.hpp"
#include "StaticVals.hpp"
#include "LogSet.hpp"
#include "conf_utl.hpp"
#include "main_cmn.hpp"

/** General Configulation data */
Config_set vdif_conf;

extern "C"{
#include <unistd.h>
#include <sys/types.h>
}
const char* version_str = "$Rev: 273 $ updated $Date: 2017-11-15 10:02:19 +0900 (水, 15 11月 2017) $ ";

void usage(){
  cerr << "Usage: " << me <<" [-p <ctrl port>] [-P <data port>] [-m <mode>} [-g <n>]\n";
  cerr << "    [-t <VDIF | RAW | K5VSSP | MK5B>] [-x <Norm|Mk5Geo>] \n"; 
  cerr << "    [-s <Stname>:<ID>][-l <logfile>] [-v <n>] [-o <Norm|Octavia>]\n";
  cerr << "  ---<Control>-----------------------------------------------\n";
  cerr << "  Server mode data receiver of VDIF/SDUP stream\n";
  cerr << " -p <port>, port number for communication with client. default=9903.\n";
  cerr << "  ---<Input>-----------------------------------------------\n";
  cerr << " -P <port>, port number for receiving data stream. default=9901.\n";
  cerr << " -m: Input data mode. ex.: 32Msps/2bit/16ch is default.\n";
  //cerr << " -o: Output data mode. ex.: 32Msps/2bit/16ch is default.\n";
  cerr << "  ---<Output>-----------------------------------------------\n";
  cerr << " -g: Number of max output threads.\n";
  //cerr << "    Supporting Output format: VDIF, VSSP32. \n";
  cerr << " -t: Output Data Frame Type [VDIF/K5VSSP/MK5B/RAW] (default: VDIF) \n" ;
  cerr << " -x: Extraction mode Normal(default) or Mk5Geo(stream 16-19/32 are \n";
  cerr << "     LSB(14,15ch).\n";
  cerr << "  ---<Information>------------------------------------------\n";
  cerr << " -s <Stname>:<ID>: Station name(8 byte) and ID(2byte) for K5/VSSP32\n";
  cerr << "     file header information. \n";
  //cerr << " -i <info file>:  Logging information(packet-loss etc.).\n";
  cerr << " -v: verbose level, 0--4 (default -1: no logging).\n";
  cerr << " -o: Octavia data.(default: Normal data)\n";
  cerr << " -l: <log file name>.\n";
  cerr << " -n: Setting frame size by number of divisions/sec.\n"; 
  cerr << " Version: " << version_str << endl;
  cerr << "-------------------------------------------------------" << endl;
  exit(1);
}


int main(int argc, char *argv[]){
  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */ 

  char logstr[256];
  int loglevel(0);
  int ndiv=10; // Number of frames in 1sec in VDIF output file
  //  for (int i = 0; i < argc; i++) { printf("%d = %s\n", i, argv[i]);  }
  int rxport(DEFAULT_DATA_PORT);
  int ctrlport(DEFAULT_CTRL_PORT2);
  //  string ip("0.0.0.0");
  //LogSet logset(me,-1);
  int ngrp(1) ;// Max number of output groups
  string inmode = "32Msps/2bit/16ch";
  //string outmode = "64Msps/2bit/4ch";
  bool mk4geo_mode = false;
  bool log_l = false;
  bool octavia_mode = false;
  string Stname("Kashim34"), Stid("kb");
  VLBI_Frame::VLBI_Frame_t frame_type(VLBI_Frame::VDIF_frm);
	
  int k=0;
  while(( k = getopt(argc, argv, "h:p:P:v:g:m:o:x:s:t:l:n:")) != -1){ 
    switch (k){
    case 'l':
      {
	Time_dat t1; t1.Set_time_to_now();
	string ydhms = t1.Print_time(13);
	sprintf(logstr,"%s%s_%s%d",me,ydhms.c_str(),optarg,getpid()); 
	
      }
      break;
    case 'p':   ctrlport  = atoi(optarg); break;
    case 'P':   rxport  = atoi(optarg); break;
    case 'm':   inmode = optarg; break;
    //case 'o':   outmode = optarg; break;
      //case 'i':   logset.Set_infofile(optarg); break;
      //case 'v':   logset.Set_level(atoi(optarg)); break;
    case 'v':   loglevel = atoi(optarg); log_l = true; break;
    case 'o':   if(strncmp(optarg,"Octavia",7) == 0) octavia_mode=true; break;
    case 'g':    ngrp = atoi(optarg); break;
    case 's':   {
      char *cp = strchr(optarg,':');
      if( cp != NULL){ *cp = '\0'; cp++; Stid = cp;}
      Stname =optarg;
      break;}
    case 'x':   if(strncmp(optarg,"Mk5Geo",6) == 0) mk4geo_mode=true; break;
    case 'n':   ndiv = atoi(optarg); break; 
    case 't':
      if(strstr(optarg,"VDIF")!=NULL){frame_type = VLBI_Frame::VDIF_frm; } 
      if(strstr(optarg,"K5VSS")!=NULL){frame_type = VLBI_Frame::K5VSSP_frm; } 
      if(strstr(optarg,"RAW") !=NULL){frame_type = VLBI_Frame::K5VSI_frm; } 
      if(strstr(optarg,"MK5B")!=NULL){frame_type = VLBI_Frame::MK5B_frm; }
      break;
    case 'h':    usage(); break;
    case '?':    cerr << "Missing following parameters\n"; usage(); break;
    case ':':    cerr << "Invalid option\n"; usage(); break;
    }
  }

  LogSet logset(logstr,-1);

  if (log_l){ logset.Set_level(loglevel); }
  logset.Start();
  
  DAS_Info   outdas;
  DAS_Info   indas;
	
  indas.SetMode(inmode);
  //outdas.SetMode(outmode);
  //if(! outdas.SetMode(outmode) || ! indas.SetMode(inmode)){
  //  cerr << "Error at mode setting\n";
  //  usage();
  //}
   
  VDIF_Frame in_frame;
  in_frame.Set_DasInfo(indas); 
  if (octavia_mode){ in_frame.Octavia_data(); }
	
  VDIFProto inp;
  inp.Set_VlbiFrame(&in_frame);
  inp.Set_DataPort(rxport);
  //  inp.Set_TransLength(length); // sec
  //  inp.Set_Speed(rate); // (Mbps)
  inp.Disp_RxRate();
  cerr << "Innet Intializing..." << flush;
  InNet_Model innet;
  cerr << " Set Protcol to Innet..." << flush;
  innet.Set_protocol(&inp);
  cerr << " Frame Set..." << flush;
  innet.Set_VlbiFrame(&in_frame);
  cerr << " DasInfo Set..." << flush;
  innet.Set_DasInfo(indas);
  cerr << endl;

  if( innet.Ready() < 0 ){
	cerr << "Error in innet.Ready()\n";
	exit(1);
  }
  
  innet.start();

  /** TxRxing Control object */ 
  TransCtl transctl_proto;
  transctl_proto.Set_CtrlPort(ctrlport);

  VLBI_Proto* outp[ngrp];   // Raw output protocol
  ///VDIF_Frame out_frame[ngrp];
  VLBI_Frame* out_frame[ngrp];
	
  for(int i=0; i<ngrp; i++){

    switch( frame_type)
    {
      case VLBI_Frame::K5VSSP_frm:
	 	out_frame[i] = new K5VSSP32_Frame; 
	 	cerr << "["<< i <<"]Output Frame: K5VSSP32 Frame\n";
	 	break;
      case VLBI_Frame::K5VSI_frm: 
		out_frame[i] = new K5VSI_Frame; 
		cerr << "["<< i <<"]Output Frame: K5VSI (Raw)Frame\n";
		break;
      case VLBI_Frame::MK5B_frm:  
		out_frame[i] = new Mk5B_Frame;
		cerr << "["<< i <<"]Output Frame: Mk5B Frame\n"; 
		break;
      case VLBI_Frame::VDIF_frm:
	    default:
		out_frame[i] = new VDIF_Frame; 
		out_frame[i]->Set_Nframe_sec( ndiv );
		cerr << "["<< i <<"]Output Frame: VDIF Frame\n";
		break;
    }
  	
  	if (octavia_mode){ out_frame[i]->Octavia_data(); }
  	
    outp[i] = new RAW_Proto;
	out_frame[i]->Set_DasInfo(outdas); 
	outp[i]->Set_VlbiFrame(out_frame[i]); 
	out_frame[i]->Set_StnameID(Stname.c_str(), Stid.c_str());

  	///if(frame_type == VLBI_Frame::VDIF_frm){
	///  	out_frame[i].Set_Nframe_sec(ndiv);
	///  	cerr << "vlbi Div=" << out_frame[i].Nframe_sec() << endl;
  	///}
  	///out_frame[i].Set_Nframe_sec( ndiv ); 
  	///cerr << "vdif Div=" << out_frame[i].Nframe_sec() << endl;
  	
    if(mk4geo_mode){
      outp[i]->SetExmode_Mk4Geo(); // Extracting LSB 8,9ch->15,16ch
      cerr << "I: Rcving mode["<< i << "] Mk4VSI-Geodesy" << endl;
    }
  }

  transctl_proto.Set_ApplProto(outp, ngrp); // Set TransInfo
  //  transctl_proto.Set_ApplProto(&outp[0], ngrp); // Set TransInfo
  transctl_proto.Set_InputDev(&innet);
  if(! transctl_proto.SetupCheck()) exit(1);

  transctl_proto.start(); // Transaction server starting
	
//   while(1){
//     Logsystem::info_logger()->Tstamp()
//       << transctl_proto.DumperState() << "\n";
//     Logsystem::info_logger()->Flush();
//     sleep(1);
//   }

  transctl_proto.join();  // waiting for ending

  inp.Disp_RateOff();  

  for(int i=0; i < ngrp; i++){ delete outp[i]; }
  logset.Stop();
}
