/** @file  MK5B_Frame.hpp 
 * A interface Class for MK5B Frame.
 * @Version $Rev: 263 $
 * @Author M.Sekido
 * @Date  $Date: 2017-11-11 15:39:57 +0900 (土, 11 11月 2017) $
 * $Id: Mk5B_Frame.hpp 263 2017-11-11 06:39:57Z vlbi $ 
 */

#ifndef __MK5B_Frame_HPP__
#define __MK5B_Frame_HPP__

#include "Mk5B_Header.hpp"
#include "VLBI_Frame.hpp"


//========== Mk5B ===from Here  ====================
class Mk5B_Frame: public VLBI_Frame{
  //  K5_VSI *k5vsi_dev;  //< K5_VSI Device  
  Mk5B_Hdr m5hdr;  /// Mark5B header
  CRC_mdl*  pCRC_mdl;
  FILE *fp;  ///< File pointer for reading VDIF file.
  /** Number of frames in 1 sec. 
   * Set in read_every_sec_hdr()  **/
  // uint32_t  nframe_sec; comment out 2014/5/22 miya, nframe_sec moved to VLBI_frame
  /// Buffer for frequent reading data in Read_frame().
  vector <uint8_t> workbuf; 
public:
  /// Default Constructor 
  Mk5B_Frame():VLBI_Frame(DEFAULT_MK5B_FRAMESIZE),fp(0){ nframe_sec=0; /// nframe_sec(0) del 2014/5/22 miya
    pCRC_mdl = new CRC_jive;}
  ~Mk5B_Frame(){delete pCRC_mdl;} ///< Destructor 
  /// Return Frame type
  const VLBI_Frame_t Type()const{return MK5B_frm;}
  string TypeName(){return string("Mk5B");}
  /// Decomposing frame 
  //  int Decompose_frame(const uint8_t* buf, uint32_t size );
  /// Writing Frame 
  int Make_frame(const uint8_t* pkt,  const uint8_t* src, const int data_size );
  int Make_framezero(const uint8_t* pkt ); // add 2016/9/29 miya
  /// Framesize (bytes) including header 
  //  uint32_t Framesize()const{return m5hdr.Framesize();};
  /// Headersize (bytes)
  int Headsize()const{return Mk5B_Hdr_size;}
  int Set_header(const uint8_t* buf, const int ifrm=-1);
  void Reset_header(){m5hdr.iframe=0;} ///< Reset sequential number of header
  void Cntup_ifrm(){m5hdr.iframe++;} // add miya 2016/11/9
  int Convert();
  /// Set Mark5 header
  //   void Set_mark5header( uint32_t *header, 
  // 			const time_t epoch, const double sec, 
  // 			const uint16_t iblk);
  const int Set_DasInfo(const DAS_Info& ds); //add miya 2017/5/9
  double sec_per_frame; // add miya 2017/5/9
  //  void Set_Framesize(const uint32_t sz){;}
  /// Dummy function of setting Station-name and Station-ID for Mk5B
  void Set_StnameID(const char* name, const char* id=0){;}
  /// Dummy function of Retruning Stname;
  void Get_Stname(string & name){;}
  /// Dummy Function of Retruning Station ID
  void Get_StnID(string & id){;}
  /// Open for reading;
  const bool ROpen(string fname){
    fp = fopen(fname.c_str(), "r"); 
    if (fp == NULL) return false; else return true;
  }
  /**! Read data in unit of seconds
   *  \param sec Time duration for reading data.
   *  \return number of frames, could be read.
   **/
  const long Read_sec(vector <uint8_t>& rawbuf, const int sec=1);
  /**! Read data in unit of frames
   *  \param hdarr header array to store the header data
   *  \param rawbuf 
   *  \return number of read frames.
   **/
  const long Read_frame(vector <Mk5B_Hdr>& hdrarr, vector <uint8_t>& rawbuf,
			const int frame=1);
  /**! Read all header of every second tics and returns its array
   *  \return Total Frame number in the data.
   **/
  const long Read_every_sec_hdr(vector <Mk5B_Hdr>& hdrarr, const size_t nsec=0);
  /**! Read all header of every frames and returns its array
   *  \return Total Frame number in the data.
   **/
  const long Read_every_frame_hdr(vector <Mk5B_Hdr>& hdrarr, 
		      const size_t nfrm=100, const bool rewind_flag=false);
  /// Rewind and move to first header point in file pointer for reading.
  const bool Rewind(void);
  const long Nframe_per_sec(void)const {return nframe_sec;}
  
  const bool Fix_nFrame_sec(void);
};

inline const bool Mk5B_Frame::Fix_nFrame_sec(void){
  //  Mk5B_Hdr hdr;
  char buf[framesize];
  Mk5B_Hdr *phdr = (Mk5B_Hdr *)buf;
  long offset(0);
  char exceptionMessage[1024];

  try{
    rewind(fp);
    if ( ! Mk5B_Hdr_find(fp, offset, *phdr) ){
      strcpy(exceptionMessage," Failed to find Header");
      throw exceptionMessage;
    }

    // Skip to the next exact second.
    while( phdr->iFrame() >0){
      if(fread(buf,framesize,1,fp) <1){
	strcpy(exceptionMessage," Failed to find First sec.");
	throw exceptionMessage;
      };
      if(!phdr->HD_Match()){
	strcpy(exceptionMessage," Missed Hdr in finding First sec.");
	throw exceptionMessage;
      }
      offset += framesize;
    }

    // int sod[2];
    // sod[0] = (int)phdr->Sod();    
    uint32_t iframe[3] ={0,0,0};
    iframe[0] = phdr->iFrame();
    //    fseek(fp,framesize,SEEK_CUR);
    //  int32_t hdsize = head.Headsize();
    //  int bufsize = hdsize * 10;
    while( phdr->iFrame() ==0){
      fread(buf,framesize,1,fp);
    }
    // Skip to the next exact second.
    int icnt=1;
    while( phdr->iFrame() >0){
      icnt++;
      if(phdr->HD_Match()){
	iframe[1] = phdr->iFrame();
      }else{
	strcpy(exceptionMessage," Missed Hdr in finding Next sec.");
	throw exceptionMessage;
      }
      if(fread(buf,framesize,1,fp) <1){
	strcpy(exceptionMessage," Failed Fread() in finding Next sec.");
	throw exceptionMessage;
      };
    }
    nframe_sec = icnt;
    if(nframe_sec != (iframe[1]+1)){
      fprintf(stderr,
	      "W Mk5B_Frame.Fix_nFrame_sec():nframe_sec(%d)-1 != iframe_end(%d) - iframe_first(%d)\n",
	      nframe_sec, iframe[1], iframe[0]);
    }

    fseek(fp,offset,SEEK_SET); // Rewinding to the first header

    return true;
  } // End try

  catch(char* str){
    cerr << "E:Mk5B_Frame.Fix_nFrame_sec():" << str << endl;
    return false;
  }
  
}

inline int Mk5B_Frame::Make_frame(const uint8_t* pkt,  const uint8_t* src,
				   const int data_size){
  Set_header(pkt);
  pkt += Mk5B_Hdr_size;
  memcpy((void*) pkt, src, data_size );
  return data_size + Mk5B_Hdr_size;
}

inline int Mk5B_Frame::Make_framezero(const uint8_t* pkt ){ 
  Set_header(pkt); return 0;} // add 2016/9/29 miya

inline const bool Mk5B_Frame::Rewind(void){
  long offset(0); 
  Mk5B_Hdr hdr;
  rewind(fp);
  if(! Mk5B_Hdr_find(fp, offset, hdr) ) return false;
  fseek(fp,offset,SEEK_SET);
  return true;
}

//add miya 2017/5/9 start---------------------------------------------
inline  const int Mk5B_Frame::Set_DasInfo(const DAS_Info& ds){
  das_inf = ds;
  sec_per_frame = ds.Bytes_sec();
  printf("Mk5B_Frame ds.Bytes_sec %d\n", ds.Bytes_sec() );
  sec_per_frame = ( framesize - Mk5B_Hdr_size )  / sec_per_frame;
  return 0;
}
//add miya 2017/5/9 end---------------------------------------------

//========== Mk5B ===to   Here  ====================

#endif
