#ifndef __Mk5B_HEADER__
#define __Mk5B_HEADER__
/** Mk5B Header Class library
 * @Version $Rev: 220 $
 * @Author M.Sekido
 * @Date  $Date: 2017-05-25 14:06:42 +0900 (木, 25  5月 2017) $
 * $Id: Mk5B_Header.hpp 220 2017-05-25 05:06:42Z vlbi $ 
 */

// #include "VLBI_Header.hpp"

#include <string>
#include "time_lib.hpp"
using namespace std;


//========== Mk5B ===From here ====================
const int Mk5B_Hdr_size = 16;  // (bytes)
const int DEFAULT_MK5B_FRAMESIZE = 10016;

struct Mk5B_Hdr{
  uint32_t sync_word; ///< sync word ffff
  unsigned iframe:15;  ///<  Frame Number in that sec
  unsigned tvg_flag:1; ///< tvg_flag
  uint16_t user_dat; ///< user data
  unsigned s0:4;  ///< sod0 BDC
  unsigned s1:4;  ///< sod1 BDC
  unsigned s2:4;  ///< sod2 BDC
  unsigned s3:4;  ///< sod3 BDC
  unsigned s4:4;  ///< sod4 BDC
  unsigned mjd0:4;  ///< mjd0 BDC
  unsigned mjd1:4;  ///< mjd1 BDC
  unsigned mjd2:4;  ///< mjd2 BDC
  uint16_t crc;   ///< CRC code 
  unsigned ms4:4;  ///< Below second BDC
  unsigned ms3:4;  ///< Below second BDC
  unsigned ms2:4;  ///< Below second BDC
  unsigned ms1:4;  ///< Below second BDC
public:
  /// Constructor 
  Mk5B_Hdr(){memset(this,0,Mk5B_Hdr_size); sync_word=0xffffffff;}
  double Sod()const; ///< Return total Sec of the day
  /// Print Header information 
  void Print_hd(FILE *fp=stdout){fprintf(fp,"%s\n",Str_hd().c_str());}
  /// Returns Header infomation by string 
  string Str_hd(void);
  /// Return framesize number 
  int iFrame()const {return (int)iframe;}
  /// Returns frame size (byte)
  int Framesize() {return DEFAULT_MK5B_FRAMESIZE;} 
  int Headsize()const {return (Mk5B_Hdr_size);} ///< return header size (bytes)
  Time_dat Epoch(); ///< return Time_dat data of the epoch
  /// Return integer MJD of the epoch
  uint32_t Epoch_imjd(){return (uint32_t)(floor(Epoch().Mjd()));}
  double Epoch_sod(){return Sod();} ///< return sec of the day (sec)  
  /// Returns test sync_word == 0xABADDEED
  bool HD_Match(){return  sync_word == 0xABADDEED ;}
  void Set_header(Time_dat& t1, double sec_per_frame); ///< Setup header data with time //add double sec_per_frame miya 2017/5/9
};

inline double  Mk5B_Hdr::Sod() const{
  double sod = (double)(((s4*10+s3) *10 + s2)*10 + s1)*10 + s0;
  sod += (double)((( ms4*0.1 + ms3)*0.1 + ms2)*0.1 + ms1)*0.1;
  return sod;
}

Mk5B_Hdr*   Mk5B_Hdr_find(const uint8_t*  buf, 
			  const uint32_t nbyte, long& offset);
const bool  Mk5B_Hdr_find(FILE *fp, long& offset, Mk5B_Hdr& hdr);

#endif
