/** 
 * @Version $Rev: 277 $
 * @Author M.Sekido
 * @Date  $Date: 2018-01-31 17:07:23 +0900 (水, 31  1月 2018) $
 * $Id: Raw_Proto.cpp 277 2018-01-31 08:07:23Z vlbi $ 
 */

#include "Raw_Proto.hpp"

/** TX1sec -- Trasnmission in single stream, vector version -- */
int32_t RAW_Proto::TX1sec(time_t t, vector <uint8_t>& arr){
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): started.\n";
  Logsystem::debug_logger()->Flush();
    
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  if(  is_network && addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): host/port is not set.\n";
    return -1;
  }

  // Rset Frame counter for VDIF
  if(vlbi_frame->Type() == VLBI_Frame::VDIF_frm ||
     vlbi_frame->Type() == VLBI_Frame::MK5B_frm ) vlbi_frame->Reset_header();

  //  status = SNDNG;
  int32_t remain = arr.size();
  uint8_t* frame_ptr = &packet[0];
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  uint8_t* dat_ptr = &arr[0];
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize; 

  if( remain % pld_size != 0){
    fprintf(stderr,"E @RAW_Proto::TX1sec(): bytesec(%u) mod pld(%d) !=0\n",
	    remain, pld_size);
    cerr << "Frame(Payload) size must be ajusted appropriately" << endl;
    return -1;
  }
  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize << "\n";
    return -1;
  }

  int npkt = remain / pld_size;
  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);

  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  for(int ipkt=0; ipkt < npkt && err_cnt>0; ipkt++){
    //    int packed_byte = vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    dat_ptr += pld_size;   
    
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if( nwrote < (int)pktsize){ 
	if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      }
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
      if(nwrote < pktsize){ 
	if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stoppd.";
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      }
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
    
    if(ipkt != 0 && (ipkt % rate_adj_intvl == 0) ) Rate_Adj();
  }
  
  if( remain !=0){ 
    Logsystem::error_logger()->Tstamp() 
      << "E @RAW_Proto::TX1sec():  remian=" << remain  << " bytes \n";
    if(remain > 0) return -remain;
    else           return -1;
  }

  Logsystem::debug_logger()->Tstamp() 
    << "RAW_Proto::TX1sec(): " << total_sent  << " Bytes sent, end.\n";
  Logsystem::debug_logger()->Flush();

  return total_sent;
}


/** TX1secSub -- Trasnmission of sub-channels of data, vector version -- */
int32_t 
RAW_Proto::TX1secSub(const time_t t1, vector <uint8_t>& arr, const DAS_Info* indas){
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1secSub(): started.\n";
  Logsystem::debug_logger()->Flush();
    
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  try{

    if(  is_network && addr.sin_addr.s_addr == 0){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1sec(): host/port is not set.\n";
      return -1;
    }
    // Rset Frame counter for VDIF
    if(vlbi_frame->Type() == VLBI_Frame::VDIF_frm ||
       vlbi_frame->Type() == VLBI_Frame::MK5B_frm ) vlbi_frame->Reset_header();

    //  int32_t remain = ExtSubCh(arr,das); // Extracting data for subchannel
    int ratio = indas->Speed()/vlbi_frame->DAS_Inf().Speed();
    int32_t remain = arr.size()/ratio;

    uint8_t* frame_ptr = &packet[0];
    //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
    //  uint8_t* dat_ptr = &preSendBuf[0];
    int framesize = vlbi_frame->Framesize();
    int pld_size = framesize - vlbi_frame->Headsize();
    int pktsize = framesize; 

    if( remain % pld_size != 0){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1sec(): bytesec("<< remain
	<< ") mod pld(" << pld_size << ") !=0\n"
	<< "Frame(Payload) size must be ajusted appropriately.\n";
      return -1;
    }

    if( (int)packet.size() < pktsize ){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1sec(): Inconsistent packet_size:"
	<< (int)packet.size() << " != protocol size:" << pktsize << "\n";
      return -1;
    }
    vector <uint8_t> databuf(pld_size);
    uint8_t* dat_ptr = &databuf[0];
    int npkt = remain / pld_size;
    //  int rate_adj_intvl = npkt / rate_adj_frq;
    const int err_lmt = ERROR_CNT_LIMIT;
    int nwrote(0);
    int err_cnt=err_lmt;  
    int32_t total_sent(0);

    extch.Setup(&arr[0], arr.size(), indas, vlbi_frame->DAS_Inf());
    bool Is1bitext = indas->Qbit() ==2 && vlbi_frame->DAS_Inf().Qbit() ==1;
    //  int pkt_cnt=0;
    //  while(pbuf < arr.end()){
    int32_t nbyte(0);
    for(int ipkt=0; ipkt < npkt && err_cnt > 0; ipkt++){
      if(Is1bitext)  nbyte = extch.Extract1bit_from2bit(databuf, pld_size);
      else         nbyte = extch.Extract(databuf, pld_size);
      if( nbyte < 0 ){
	cerr << "E @RAW_Proto::Tx1secSub(): Failed extch.Extract()"
	     << " err_cnt=" << err_cnt << endl;
	err_cnt--; 
	if(nbyte < -1) err_cnt = -1;
	continue;
      }
      vlbi_frame->Make_frame(frame_ptr, dat_ptr, nbyte);
      //    dat_ptr += pld_size;   
    
      switch (transLayerProtocol){
      case TCP_IO:
      case FILE_IO: // Writing only Payload part without seq number.
	nwrote = _sock->send(&packet[0], pktsize);
	if( nwrote < (int)pktsize){ 
	  if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
	  fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	  if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
	}
	break;
      case UDP_IO:
	nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			       (char*)&packet[0], pktsize);
	if(nwrote < pktsize){ 
	  if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
	  fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	  if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
	}
	break;
      }
      rate_meas.Add_Snd_Pld(pld_size);
      total_sent += nwrote;
      remain -= pld_size;
      err_cnt = err_lmt;      
    }
  
    if( remain !=0){ 
      Logsystem::error_logger()->Tstamp() 
	<< "E @RAW_Proto::TX1sec():  remian=" << remain 
	<< " bytes.\n";
      if(remain > 0) return -remain;
      else           return -1;
    }

    Logsystem::debug_logger()->Tstamp()
      << "RAW_Proto::TX1sec(): " << total_sent << " Bytes sent, end.\n";
    Logsystem::debug_logger()->Flush();
    
    return total_sent;
  }

  catch (const char* str){
    Logsystem::error_logger()->Tstamp() << str << " \n";
      
    return -1;
  }

}


/** TX1sec -- Trasnmission in single stream, pointer version -- */
int32_t RAW_Proto::TX1sec(time_t t, const uint8_t* arr, const int32_t ndat) {
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1sec(): started.\n";
  Logsystem::debug_logger()->Flush();
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  if(  is_network && addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): host/port is not set.\n";
    return -1;
  }
  // Rset Frame counter for VDIF
  if(vlbi_frame->Type() == VLBI_Frame::VDIF_frm ||
     vlbi_frame->Type() == VLBI_Frame::MK5B_frm ) vlbi_frame->Reset_header();

  //posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED); // move 2016/9/13 miya
  //  status = SNDNG;
  int32_t remain = ndat;
  const uint8_t* frame_ptr = &packet[0];
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
  const uint8_t* dat_ptr = arr;
  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize; 
  uint32_t datasyncsize(0); // add miya 2016/9/13
  //Rot_bar rot;// Rotation progress bar
  if( remain % pld_size != 0){
    fprintf(stderr,"E @RAW_Proto::TX1sec(): bytesec(%u) mod pld(%d) !=0\n",
	    remain, pld_size);
    cerr << "Frame(Payload) size must be ajusted appropriately" << endl;
    return -1;
  }
  
  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1sec(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize  << "\n";
    return -1;
  }

  int npkt = remain / pld_size;
  //  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0); 
  int err_cnt=err_lmt;  
  int32_t total_sent(0);

  //  int pkt_cnt=0;
  //  while(pbuf < arr.end()){
  for(int ipkt=0; ipkt < npkt && err_cnt >0; ipkt++){
  
    //    int packed_byte = vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
  	vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
    dat_ptr += pld_size;   

    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
      nwrote = _sock->send(&packet[0], pktsize);
      if(datasyncsize>DISK_SYNC_SIZE){ // add 2016/9/13 miya
      	datasyncsize=0; // add 2016/9/13 miya
	fdatasync(_sock->get_sockd());// add 2014/10/8 miya
      	posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED);
      }else{ datasyncsize += pktsize;} // add 2016/9/13 miya
      //posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED); // add 2014/10/8 miya
      if( nwrote < (int)pktsize){
	if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      }
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
      if(nwrote < pktsize){ 
		if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
		fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
		if(nwrote <0){perror("E @TX1sec():"); err_cnt--; continue;}
      }
      break;
    }
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
    //if(ipkt % 5000 == 0) rot.Step(); // Rotating bar 
  }

  if( remain !=0){ 
    Logsystem::error_logger()->Tstamp() 
      << "E @RAW_Proto::TX1sec():  remian=" << remain   << " bytes.\n";
    if(remain > 0) return -remain;
    else           return -1;
  }

  Logsystem::debug_logger()->Tstamp() 
    << "RAW_Proto::TX1sec(): " << total_sent << " Bytes sent, end.\n";
  Logsystem::debug_logger()->Flush();
  return total_sent;
}

/** TX1seczero --Writing ZERO single stream, pointer version -- */ // add 2016/9/29 miya
int32_t RAW_Proto::TX1seczero(time_t t,  const int32_t ndat){
  Logsystem::error_logger()->Tstamp() << "RAW_Proto::TX1seczero(): called.\n";
  Logsystem::error_logger()->Flush();
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);
  if(  is_network && addr.sin_addr.s_addr == 0){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1seczero(): host/port is not set.\n";
    return -1;
  }
  // Rset Frame counter for VDIF
  if(vlbi_frame->Type() == VLBI_Frame::VDIF_frm ||
     vlbi_frame->Type() == VLBI_Frame::MK5B_frm ) vlbi_frame->Reset_header();

  //posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED); // move 2016/9/13 miya
  //  status = SNDNG;
  int32_t remain = ndat;
  uint8_t* frame_ptr = &packet[0];
  //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();

  int framesize = vlbi_frame->Framesize();
  int pld_size = framesize - vlbi_frame->Headsize();
  int pktsize = framesize; 
  uint32_t datasyncsize(0);
  uint8_t *buffer;
  buffer=(uint8_t *)malloc(remain);
  memset(buffer,0x00,pld_size);
  const uint8_t* dat_ptr = buffer;
 // const uint8_t* dat_ptr = arr;
  if( remain % pld_size != 0){
    fprintf(stderr,"E @RAW_Proto::TX1seczero(): bytesec(%u) mod pld(%d) !=0\n",
	    remain, pld_size);
    cerr << "Frame(Payload) size must be ajusted appropriately" << endl;
    return -1;
  }
  
  if( (int)packet.size() < pktsize ){
    Logsystem::error_logger()->Tstamp()
      << "E @RAW_Proto::TX1seczero(): Inconsistent packet_size:"
      << (int)packet.size() << " != protocol size:" << pktsize  << "\n";
    return -1;
  }

  int npkt = remain / pld_size;
  //  int rate_adj_intvl = npkt / rate_adj_frq;
  const int err_lmt = ERROR_CNT_LIMIT;
  int nwrote(0); 
  //int hwrote(0); 
  //int nseek(0); 
  //uint8_t z(0);
  int err_cnt=err_lmt;  
  int32_t total_sent(0);

  vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size); // add miya 2016/11/9
	
  for(int ipkt=0; ipkt < npkt && err_cnt >0; ipkt++){
  	///vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);
  	///dat_ptr += pld_size;   
    ///vlbi_frame->Make_framezero(frame_ptr);
  	///dat_ptr += vlbi_frame->Headsize();   
    switch (transLayerProtocol){
    case TCP_IO:
    case FILE_IO: // Writing only Payload part without seq number.
    	nwrote = _sock->send(&packet[0], pktsize);
      //hwrote = _sock->send(&packet[0], vlbi_frame->Headsize());
      //nseek = _sock->seek(pld_size-1);
      //nwrote = _sock->send(&z, 1);
      //nwrote =nwrote+hwrote;
    	
      if(datasyncsize>DISK_SYNC_SIZE){ // add 2016/9/13 miya
      	datasyncsize=0; 
	    fdatasync(_sock->get_sockd());
      	posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED);
      }else{ datasyncsize += pktsize;} 
      if( nwrote < (int)pktsize){
	  	if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
	 	fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
		if(nwrote <0){perror("E @TX1seczero():"); err_cnt--; continue;}
      }
      break;
    case UDP_IO:
      nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			     (char*)&packet[0], pktsize);
      if(nwrote < pktsize){ 
		if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
		fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
		if(nwrote <0){perror("E @TX1seczero():"); err_cnt--; continue;}
      }
      break;
    }
  	//vlbi_frame->Cntup_ifrm();  // add miya 2016/11/9
  	vlbi_frame->Make_framezero(frame_ptr);  // add miya 2016/11/9
    rate_meas.Add_Snd_Pld(pld_size);
    total_sent += nwrote;
    remain -= pld_size;
    err_cnt = err_lmt;      
  }
  free(buffer);
  if( remain !=0){ 
    Logsystem::error_logger()->Tstamp() 
      << "E @RAW_Proto::TX1seczero():  remian=" << remain   << " bytes.\n";
    if(remain > 0) return -remain;
    else           return -1;
  }
  Logsystem::error_logger()->Tstamp() 
    << "RAW_Proto::TX1seczero(): " << total_sent << " Bytes sent, end.\n";
  Logsystem::error_logger()->Flush();
  return total_sent;
}

/** TX1secSub -- Trasnmission of sub-channels of data, vector version -- */
int32_t 
RAW_Proto::TX1secSub(const time_t t1, const uint8_t* arr, 
		     const int32_t ndat, const DAS_Info* indas){
  Logsystem::debug_logger()->Tstamp() << "RAW_Proto::TX1secSub(): started.\n";
  Logsystem::debug_logger()->Flush();
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);

  try{
    if(  is_network && addr.sin_addr.s_addr == 0){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1secSub(): host/port is not set.\n";
      return -1;
    }

  	// Rset Frame counter for VDIF
    if(vlbi_frame->Type() == VLBI_Frame::VDIF_frm ||
       vlbi_frame->Type() == VLBI_Frame::MK5B_frm ) vlbi_frame->Reset_header();

	//posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED); // move 2016/9/13 miya
  
    //  status = SNDNG;
    //  int32_t remain = ExtSubCh(arr, ndat, das); // Extracting data for subchannel
    int ratio = indas->Speed()/vlbi_frame->DAS_Inf().Speed();
    int32_t remain = ndat/ratio;

    uint8_t* frame_ptr = &packet[0];
    //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
    //  uint8_t* dat_ptr = &preSendBuf[0];
    int framesize = vlbi_frame->Framesize();
    int pld_size = framesize - vlbi_frame->Headsize();
    int pktsize = framesize; 
    uint32_t datasyncsize(0); // add miya 2016/9/13

  	if( remain % pld_size != 0){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1secSub(): bytesec("<< remain
	<< ") mod pld(" << pld_size << ") !=0\n"
	<< "Frame(Payload) size must be ajusted appropriately.\n";
      return -1;
    }

    if( (int)packet.size() < pktsize ){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1secSub(): Inconsistent packet_size:"
	<< (int)packet.size() << " != protocol size:" << pktsize << "\n";
      return -1;
    }

    vector <uint8_t> databuf(pld_size);
    uint8_t* dat_ptr = &databuf[0];
    int npkt = remain / pld_size;
    //  int rate_adj_intvl = npkt / rate_adj_frq;
    const int err_lmt = ERROR_CNT_LIMIT;
    int nwrote(0);
    int err_cnt=err_lmt;  
    int32_t total_sent(0);

    extch.Setup(arr, ndat, indas, vlbi_frame->DAS_Inf());
    bool Is1bitext = indas->Qbit() ==2 && vlbi_frame->DAS_Inf().Qbit() ==1;
 
    //  int pkt_cnt=0;
    //  while(pbuf < arr.end()){
    int32_t nbyte(0);
    for(int ipkt=0; ipkt < npkt && err_cnt > 0; ipkt++){
      if(Is1bitext)  nbyte = extch.Extract1bit_from2bit(databuf, pld_size);
      else           nbyte = extch.Extract(databuf, pld_size);
      
      if( nbyte < 0 ){
	cerr << "E @RAW_Proto::Tx1secSub(): Failed extch.Extract()"
	     << " err_cnt=" << err_cnt << endl;
	err_cnt--; 
	if(nbyte < -1) err_cnt = -1;
	continue;
      }
      vlbi_frame->Make_frame(frame_ptr, dat_ptr, nbyte);
      //    dat_ptr += pld_size;   

      switch (transLayerProtocol){
      case TCP_IO:
      case FILE_IO: // Writing only Payload part without seq number.
		nwrote = _sock->send(&packet[0], pktsize);
      if(datasyncsize>DISK_SYNC_SIZE){ // add 2016/9/13 miya
      	datasyncsize=0; // add 2016/9/13 miya
        fdatasync(_sock->get_sockd()); // add 2014/11/12 miya
        posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED); // add 2014/11/12 miya
      }else{datasyncsize=datasyncsize+pktsize;} // add 2016/9/13 miya
	if( nwrote < (int)pktsize){ 
	  if(ProtocolState()== TransInfo::STOP) throw " RawProt Stopped.";
	  fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	  if(nwrote <0){perror("E @TX1secSub():"); err_cnt--; continue;}
	}
	break;
      case UDP_IO:
	nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			       (char*)&packet[0], pktsize);
	if(nwrote < pktsize){ 
	  if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
	  fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	  if(nwrote <0){perror("E @TX1secSub():"); err_cnt--; continue;}
	}
	break;
      }
      rate_meas.Add_Snd_Pld(pld_size);
      total_sent += nwrote;
      remain -= pld_size;
      err_cnt = err_lmt;      
    }
  
    if( remain !=0){ 
      Logsystem::error_logger()->Tstamp() 
	<< "E @RAW_Proto::TX1secSub():  remian=" << remain 
	<< " bytes.\n";
      if(remain > 0) return -remain;
      else           return -1;
    }

    Logsystem::debug_logger()->Tstamp() 
      << "RAW_Proto::TX1secSub(): " << total_sent << " Bytes sent, end.\n";
    Logsystem::debug_logger()->Flush();
    return total_sent;
  }

  catch (const char* str){
    Logsystem::error_logger()->Tstamp() << str << " \n";
      
    return -1;
  }

}

/** TX1secSubzero -- Trasnmission of sub-channels of data, vector version -- */ // add 2016/9/29 miya
int32_t 
RAW_Proto::TX1secSubzero(const time_t t1, const int32_t ndat, const DAS_Info* indas){
  Logsystem::error_logger()->Tstamp() << "RAW_Proto::TX1secSubzero(): called.\n";
  Logsystem::error_logger()->Flush();
  bool is_network = (transLayerProtocol == TCP_IO 
		     || transLayerProtocol == UDP_IO);

  try{
    if(  is_network && addr.sin_addr.s_addr == 0){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1secSubzero(): host/port is not set.\n";
      return -1;
    }

  	// Rset Frame counter for VDIF
    if(vlbi_frame->Type() == VLBI_Frame::VDIF_frm ||
       vlbi_frame->Type() == VLBI_Frame::MK5B_frm ) vlbi_frame->Reset_header();

	//posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED); // move 2016/9/13 miya
  
    int ratio = indas->Speed()/vlbi_frame->DAS_Inf().Speed();
    int32_t remain = ndat/ratio;

    uint8_t* frame_ptr = &packet[0];
    //  const vector <uint8_t>::iterator pld_ptr = frame_ptr + vlbi_frame->Headsize();
    //  uint8_t* dat_ptr = &preSendBuf[0];
    int framesize = vlbi_frame->Framesize();
    int pld_size = framesize - vlbi_frame->Headsize();
    int pktsize = framesize; 
    uint32_t datasyncsize(0); // add miya 2016/9/13

  	if( remain % pld_size != 0){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1secSubzero(): bytesec("<< remain
	<< ") mod pld(" << pld_size << ") !=0\n"
	<< "Frame(Payload) size must be ajusted appropriately.\n";
      return -1;
    }

    if( (int)packet.size() < pktsize ){
      Logsystem::error_logger()->Tstamp()
	<< "E @RAW_Proto::TX1secSubzero(): Inconsistent packet_size:"
	<< (int)packet.size() << " != protocol size:" << pktsize << "\n";
      return -1;
    }

    //vector <uint8_t> databuf(pld_size);
    //uint8_t* dat_ptr = &databuf[0];
    uint8_t *buffer;
    buffer=(uint8_t *)malloc(remain);
    memset(buffer,0x00,pld_size);
    const uint8_t* dat_ptr = buffer;
  	
    int npkt = remain / pld_size;
    //  int rate_adj_intvl = npkt / rate_adj_frq;
    const int err_lmt = ERROR_CNT_LIMIT;
    int nwrote(0);
    //int hwrote(0); 
  	//int nseek(0); 
    //uint8_t z(0);
    int err_cnt=err_lmt;  
    int32_t total_sent(0);

  	vlbi_frame->Make_frame(frame_ptr, dat_ptr, pld_size);  // add miya 2016/11/9
  	
    for(int ipkt=0; ipkt < npkt && err_cnt > 0; ipkt++){
      //vlbi_frame->Make_framezero(frame_ptr);

      switch (transLayerProtocol){
      case TCP_IO:
      case FILE_IO: // Writing only Payload part without seq number.
      	nwrote = _sock->send(&packet[0], pktsize);
		//hwrote = _sock->send(&packet[0], vlbi_frame->Headsize());
		//nseek = _sock->seek(pld_size-1);
      	//nwrote = _sock->send(&z, 1);
      	//nwrote =nwrote+hwrote;
	if(datasyncsize>DISK_SYNC_SIZE){ // add 2016/9/13 miya
	  datasyncsize=0; // add 2016/9/13 miya
	  fdatasync(_sock->get_sockd()); // add 2014/11/12 miya
	  posix_fadvise(_sock->get_sockd(),0x0,0x0,POSIX_FADV_DONTNEED); // add 2014/11/12 miya
	}else{ datasyncsize += pktsize;} // add 2016/9/13 miya
	if( nwrote < (int)pktsize){ 
	  if(ProtocolState()== TransInfo::STOP) throw " RawProt Stopped.";
	  fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	  if(nwrote <0){perror("E @TX1secSubzero():"); err_cnt--; continue;}
	}
	break;
      case UDP_IO:
	nwrote = _sock->sendto((struct sockaddr*)&addr, sizeof(addr),
			       (char*)&packet[0], pktsize);
	if(nwrote < pktsize){ 
	  if(ProtocolState()== TransInfo::STOP)  throw " RawProt Stopped.";
	  fprintf(stderr, "Error: wrote %d /%d byte\n", nwrote, pktsize);
	  if(nwrote <0){perror("E @TX1secSubzero():"); err_cnt--; continue;}
	}
	break;
      }
      //vlbi_frame->Cntup_ifrm();  // add miya 2016/11/9
      vlbi_frame->Make_framezero(frame_ptr);  // add miya 2016/11/9
      rate_meas.Add_Snd_Pld(pld_size);
      total_sent += nwrote;
      remain -= pld_size;
      err_cnt = err_lmt;      
    }
    free(buffer);
    if( remain !=0){ 
      Logsystem::error_logger()->Tstamp() 
	<< "E @RAW_Proto::TX1secSubzero():  remian=" << remain 
	<< " bytes.\n";
      if(remain > 0) return -remain;
      else           return -1;
    }

    Logsystem::error_logger()->Tstamp() 
      << "RAW_Proto::TX1secSubzero(): " << total_sent << " Bytes sent, end.\n";
    Logsystem::error_logger()->Flush();
    return total_sent;
  }

  catch (const char* str){
    Logsystem::error_logger()->Tstamp() << str << " \n";
      
    return -1;
  }

}

int32_t RAW_Proto::RX_UDP(void* buf, const int32_t buf_size){
  
  int mtu = _sock->get_mtu();
  const int n=3;
  //  status = RCVNG;
  packet.resize(mtu*n);
  memset(&packet[0],0,sizeof(packet));
  memset(buf,0,buf_size);

  const int err_lmt = 10;
  int err_cnt = err_lmt;
  int framesize = vlbi_frame->Framesize();
  int32_t rcv_cnt(0);
  //  , intvl_cnt(0);
  int64_t total_read(0);
  int addr_len = sizeof(addr);
  uint8_t *b0 = (uint8_t*)buf;
  //  int32_t nseq = *pseq;
  //  int32_t npktsps = (int32_t)((double)speed*4/mtu);
  

  while( err_cnt>0 && total_read+framesize <= buf_size ){
    int nread = _sock->recvfrom((struct sockaddr&)addr, addr_len, 
			    (char*)&packet[0], mtu*n);
    if(nread == 0 ) continue;
    
    if(nread != framesize){
      fprintf(stderr, "E @recvfrom(): recevd(%d byte)/expected(%d byte)\n",
	      nread,framesize);
      err_cnt--;
      continue;
    }

    memcpy(b0, &packet[0], nread); 
    total_read += nread;
    b0 += nread;

    err_cnt=err_lmt;
    rcv_cnt++; // intvl_cnt++;

  }// while loop

  Logsystem::error_logger()->Tstamp() 
    << "I @Raw_Proto::RX(): wrote (" << total_read << " Byte).\n";
  return total_read;

}


int RAW_Proto::RXvdif(DataPage& dpage){
    cerr  << "W @Raw_Proto::RXvdif(): is not implementted yet "<< endl;
    return -1;
}

