/**  
 *  Raw_Proto.hpp 
 *  Class interface of Raw output Protocol
 * @Version $Rev: 294 $
 * @Author M.Sekido
 * @Date  $Date: 2018-06-07 15:35:19 +0900 (木, 07  6月 2018) $
 * $Id: Raw_Proto.hpp 294 2018-06-07 06:35:19Z vlbi $ 
 */


#ifndef __RAW_PROTO_HPP__
#define __RAW_PROTO_HPP__

#include <vector>
#include <string>
using namespace std;

#include "VLBI_Proto.hpp"
#include "ExtSubCh.hpp"

extern "C"{
#include "stdint.h"
}

const uint32_t DISK_SYNC_SIZE  =  100000000; //5000000; // add 2016/9/13 miya modify 2018/6/4 miya
/** 
 *  Raw_Protocol simply outputs data without any header
 *  Succeeding from VLBI_Proto class
 */

class RAW_Proto:public VLBI_Proto {
  RAW_Proto(RAW_Proto& rp){;} ///< Copy Constructor, prohibited.
public:
  /** Constructor  */
  RAW_Proto():VLBI_Proto(TransInfo::RAW_Proto, TransInfo::FILE_IO){;}
  /// Destructor
  ~RAW_Proto(){;}
  /** Transmitting single data stream with protocol */
  int32_t TX1sec(const time_t t, vector <uint8_t>& arr);
  /** Transmitting multiple  data stream with protocol */
  int32_t TX1secSub(const time_t t, vector <uint8_t>& arr, const DAS_Info* das=0);
  /**  Transmitting data from Buffer for 1sec, pointer version */
  int32_t TX1sec(const time_t t,  const uint8_t* arr, const int32_t ndat);
  /** Writing dummy ZERO */
  int32_t TX1seczero(const time_t t, const int32_t ndat); // add 2016/9/29 miya
  /// Transmitting subset channels of 1sec of data, pointer version 
  int32_t TX1secSub(const time_t t,  const uint8_t* arr, const int32_t ndat, const DAS_Info* das=0);
  /** Writing dummy ZERO */
  int32_t TX1secSubzero(const time_t t, const int32_t ndat, const DAS_Info* das=0);// add 2016/9/29 miya
  //  virtual int32_t TX1sec(time_t t, uint8_t* arr, const DAS_Info* dasinf)=0;
  int32_t RX_UDP(void* buf, const int32_t buf_size);
  int RXvdif(DataPage& dpage);
  int ApplProtoHeadSize()const{return 0;}
  void Rate_Adj(const int32_t sent_bytes=-1);
};

inline void  RAW_Proto::Rate_Adj(const int32_t sent_bytes){
  double wait = rate_meas.Rcmnd_wait();
  if(wait >0) {
    if( Logsystem::Test_Logsys(Logsystem::DEBUG, 4)){   
      Logsystem::debug_logger()->Tstamp() << "Rate_Adj(): sleep " << wait << " sec. ";
      *Logsystem::debug_logger()
	<<"Inst:" << rate_meas.Tx_Mbps()  << "Mbps (net:" << rate_meas.Tx_Pld_Mbps()
	<< "Mbps)/Target=" << rate_meas.TargetRate() << "Mbps\n";
      Logsystem::debug_logger()->Flush();
    }

// #ifdef _MY_SLEEP_
//     my_usleep(wait*1.e6);
// #else
    struct timespec req;
    req.tv_sec = (time_t)(wait);
    req.tv_nsec = (long)((wait - req.tv_sec)*1000000000);
    nanosleep(&req,0);
    // #endif
  }
}

#endif
