/**  
 * VDIFProto.hpp This file contains class interface of VDIF Protocol
 * @Version $Rev: 56 $
 * @Author M.Sekido
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (水, 03  7月 2013) $
 * $Id: VDIFProto.hpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

#ifndef __VDIFPROTO_HPP__
#define __VDIFPROTO_HPP__

#include <UDPSocket.h>
#include <vector>
#include <string>
#include <cstring>
#include <cstdio>
using namespace std;

#include "Logsystem.hpp"
#include "VLBI_Proto.hpp"
#include "VDIF_Header.hpp"
#include "algorithm"
#include "light_utl.hpp"
#include "DataPage.hpp"

extern "C" {
#include <sys/socket.h>
#include <sys/time.h>
#include <stdint.h>
}

class VDIFProto:public VLBI_Proto{
  VDIFProto(VDIFProto& vp){;} ///< Copy Constructor, prohibited.
public:
  /** Default Constructor */
  VDIFProto():
    VLBI_Proto(VDIF_Proto, UDP_IO){ }
  /** Constructor */
  VDIFProto(const int m){Set_mtu(m);}  
  /// Destructor 
  ~VDIFProto(){
    cerr << "VDIFProto:: Destructor runs.\n";
    if(rate_meas.is_running()) rate_meas.stop();
    if(_sock !=0){ delete _sock; _sock=0;}
  }
  /**  Transmitting data from Buffer for 1sec 
   * @param t Current epoch in unix time
   * @param arr vector array data.
   */
  int32_t TX1sec(const time_t t, vector <uint8_t>& arr);
  /** Transmitting subset channels of 1sec of data 
   * @param t Current epoch in unix time
   * @param arr vector array of data.
   * @param das Input DAS mode information
   */
  int32_t TX1secSub(const time_t t, vector <uint8_t>& arr, const DAS_Info* das=0);
  /**  Transmitting data from Buffer for 1sec, pointer version 
   * @param t Current epoch in unix time
   * @param arr pointer to data array
   * @param ndat size of data array in bytes 
   */
  int32_t TX1sec(const time_t t,  const uint8_t* arr, const int32_t ndat);
  /** Transmitting subset of Observation mode for 1sec of data, pointer version 
   * @param t Current epoch in unix time
   * @param arr pointer to data array
   * @param ndat size of data array in bytes 
   * @param das Input DAS mode information
    */
  // TX1seczero add 2016/9/29 miya
  int32_t TX1seczero(const time_t t, const int32_t ndat){
    fprintf(stderr, "VDIFProto::TX1seczero() is not necessary,\n");
    return -1;
  }
  /** Transmitting subset of Observation mode for 1sec of data, pointer version 
   * @param t Current epoch in unix time
   * @param arr pointer to data array
   * @param ndat size of data array in bytes 
   * @param das Input DAS mode information
   */
  int32_t TX1secSub(const time_t t,  const uint8_t* arr, const int32_t ndat, const DAS_Info* das=0);
  // TX1secSubzero add 2016/9/29 miya
  int32_t TX1secSubzero(const time_t t,   const int32_t ndat, const DAS_Info* das=0){
    fprintf(stderr, "VDIFProto::TX1secSubzero() is not necessary,\n");
    return -1;
  }
  //   /** @doc Receiving data to Buffer for 1 sec */
  //   int32_t RX1sec(vector <uint8_t>& arr);
  /** Transmitting data from File
   * @param fp: File pointer as data source
   * @return size of data transmitted.
   */
  int32_t TX(FILE *fp);
  /**  Receiving data to File */
  //  int32_t RX(FILE *fp=NULL);
  /**  Receiving data to Buffer */
  int32_t RX_UDP(void* buf, const int32_t buf_size);
  ///int RX_SUDP(DataPage& dpage);
  int RX_VDIF(DataPage& dpage);
	///const uint64_t RX_SUDP(FILE* fp);
  const uint64_t RX_VDIF(FILE* fp);
	int RXvdif(DataPage& dpage);
  //  ApplProtocol_t ApplProtocol_Type()const {return SUDP_Proto;}
  int ApplProtoHeadSize()const{return 0;} ///VDIF_HEAD_SIZE
  TransInfo::ApplProtocol_t ApplProtocol_Type()const {
    return TransInfo::VDIF_Proto;}
  //  void Set_VlbiFrame(VLBI_Frame * vf);
  void Rate_Adj(const int32_t sent_bytes=-1);
  //  void Runflag_stop(){runflag = false; cerr << "runflag = false" << endl; }
  bool Tx_Ready(){ return VLBI_Proto::Tx_Ready(); } ///vdifhdr.ResetSeq(); 
  void FirstPacket(); ///< Senidng INIT cmd packet 
  void LastPacket();  ///< Senidng END cmd packet 
  void Finish();      ///< Reset EndEpoch and StopTimer
private:
  static const bool Send_Enable=true; // add miya 2015/6/23
  static const bool Ack_Wait_Call=true; // add miya 2015/6/23
};


inline void  VDIFProto::Rate_Adj(const int32_t sent_bytes){
  double wait = rate_meas.Rcmnd_wait();
  if(wait >0) {
    if( Logsystem::Test_Logsys(Logsystem::DEBUG, 4)){   
      Logsystem::debug_logger()->Tstamp() << "Rate_Adj(): sleep " << wait << " sec. ";
      *Logsystem::debug_logger()
	<<"Inst:" << rate_meas.Tx_Mbps()  << "Mbps (net:" << rate_meas.Tx_Pld_Mbps()
	<< "Mbps)/Target=" << rate_meas.TargetRate() << "Mbps\n";
      Logsystem::debug_logger()->Flush();
    }

#ifdef _MY_SLEEPVDIF_
    my_usleep2_vdif((ulong)(wait*1.e6));
#else
    struct timespec req;
    req.tv_sec = (time_t)(wait);
    req.tv_nsec = (long)((wait - req.tv_sec)*1000000000);
    nanosleep(&req,0);
#endif
  }
}


inline double tim_vdif(void) {
  struct timeval tv;
  double t;

  gettimeofday(&tv, NULL);
  t = (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;
  return t;
}

inline void my_usleep_vdif(double usec) {
  double now, till;
  //  int n;
  /* The current time */
  now = tim_vdif();

  /* The time we are sleeping till */
  till = now+usec/1.0e6;
  /* and spin for the rest of the time */
  //  n = 0;
  while (now<till) {
    now = tim_vdif();
    //    n++;
  }
  
}

inline const ulong Tdif_vdif(const struct timeval tm1, const struct timeval tm2){
  ulong dt = (tm2.tv_sec - tm1.tv_sec)*1000000 + (tm2.tv_usec - tm1.tv_usec);
  return dt;
};

inline void my_usleep2_vdif(const ulong usec){
//   struct timeval tm2(tm1);
//   tm2.tv_usec += usec;
//   if(tm2.tv_usec > 1000000){tm2.tv_sec++; tm2.tv_usec -= 1000000;}
  
  struct timeval tm1;
  gettimeofday(&tm1,0);
  struct timeval tm;
  do{
    gettimeofday(&tm,0);
  }while( Tdif_vdif(tm1,tm) < usec);

}

inline void my_usleep2_vdif(const ulong usec,const struct timeval tm1){
//   struct timeval tm2(tm1);
//   tm2.tv_usec += usec;
//   if(tm2.tv_usec > 1000000){tm2.tv_sec++; tm2.tv_usec -= 1000000;}
  
  struct timeval tm;
  do{
    gettimeofday(&tm,0);
  }while( Tdif_vdif(tm1,tm) < usec);

}

#endif
