/** 
 * @Version $Rev: 269 $
 * @Author M.Sekido
 * @Date  $Date: 2017-11-14 21:37:16 +0900 (火, 14 11月 2017) $
 * $Id: VDIF_Frame.cpp 269 2017-11-14 12:37:16Z vlbi $ 
 */


#include <cstring>
#include "VDIF_Frame.hpp"
extern "C"{
#include <errno.h>
#include <arpa/inet.h>
}

extern const long ELECS_NO=102300; // Cycle of +128byte in case fo elecs data

/******************************************/
VDIF_Frame::VDIF_Frame():VLBI_Frame(DEFAULT_MTU_SIZE),_fp(0),
			 _workbuf(0),_hdr_tested(-1),
			 _frame_slip_margin(Frame_Slip_Margin_DEF),elecs_flag(false){
    Time_dat t1(2000,1,1,0,0,0);
    //    _ref_epoch0 = t1;
    _vdif_hdr.Set_Nch(das_inf.Nch());
    _vdif_hdr.Set_Qbit(das_inf.Qbit());
    _vdif_hdr.Set_Srate(das_inf.Srate()/1000);
}

/* Return Numb of Frames and Intial Headr of the second *********/
long VDIF_Frame::Read_1sec_hdr(vector <VDIF_Hdr>& hdrarr, const long nread,
			       const int year){
  if(_fp==0){
    cerr << "E Failed VDIF_Frame::Read_1sec_hdr(): File is not opened yet\n";
    return -1;
  }
  int yr = year;
  long offset[2]={0,0}; /// First reading Header
  if(! VDIF_Hdr_find(_fp, offset[0], &_vdif_hdr)){
    if( ! EndofFile_Yes()){
      cerr << "E VDIF_Frame::Read_1sec_hdr(): Failed to find VDIF-Header." 
	   << endl;
      perror("E VDIF_Frame::Read_1sec_hdr(): ");
    }
    return -1;
  }; 
  uint8_t edv = Hdr().EDV();
  // The first header pos in the file.
  if( _offset_0 < 0) _offset_0=offset[0];

  _hdr_tested = 0; ///< Set test-flag to 'OK'
  long hdrsize = _vdif_hdr.Headsize();
  framesize = _vdif_hdr.Framesize();
  char *dummy = new char[framesize];
  VDIF_Hdr* phdr = (VDIF_Hdr*)dummy;
  if( (long)hdrarr.size() < nread ) hdrarr.resize(nread);

  //  int sec_cnt=0;
  //  int nerr=10;
  //  nframe_sec =0;
  Time_dat t1 = _vdif_hdr.Epoch();
  Time_dat t2=t1;
  int nframe_max=0;
  int icnt=0;
  long hdr_pos[2];
  long iframe0=_vdif_hdr.iFrame();
  long iframe1=iframe0;
  long nframe_sec1 = Nframe_sec();
  //  long skip_frame=nframe_sec1 - iframe;
  bool jump_end=false;

  while( t2 == t1 && ! EndofFile_Yes()){
    //while(skip_frame >0 && iframe1 >= iframe0){
    if( fread(dummy, hdrsize,1,_fp) != 1 ){
      //      cerr << "E VDIF_Frame::Read_1sec_hdr(): Failed to read.\n";
      perror("E VDIF_Frame::Read_1sec_hdr(): ");
      jump_end=true;
      break;
    }
    fseek(_fp, -hdrsize,SEEK_CUR);
    if( (phdr->iFrame() <= nframe_sec1 ) &&
	( (year == 0 && phdr->HD_Match() && edv == phdr->EDV() ) ||
	  (year > 0 && phdr->HD_Match( "Y", &yr)) ) ) {
      iframe0=iframe1;
      iframe1=phdr->iFrame();
      if (icnt < nread) hdrarr[icnt] = *phdr;
      if(nframe_max < Hdr().iFrame()) nframe_max =  phdr->iFrame();
      t2 =phdr->Epoch();
      icnt++;
      
    }else{
      if(verbose_no >= _VERB_WARN )
	cerr <<  "W VDIF_Frame::Read_1sec_hdr(): next Frame at expected position."    << endl;
      hdr_pos[0] = ftell(_fp);
      if( ! find_hdr()){
	if( ! EndofFile_Yes()){
	  cerr << "E VDIF_Frame::Read_1sec_hdr(): Failed to find header."     << endl;
	  perror("E VDIF_Frame::Read_1sec_hdr():");
	  jump_end=true;
	}
	break;
      }else{
	hdr_pos[1] = ftell(_fp);
	if(verbose_no >= _VERB_INFO )
	  fprintf(stderr,
		  "I Read_1sec_hdr():Headr found at offset:%ld Byte.\n", 
		  hdr_pos[1] - hdr_pos[0]);
	iframe0=iframe1;
	iframe1=phdr->iFrame();
	if (icnt < nread) hdrarr[icnt] = Hdr();
	if(nframe_max < Hdr().iFrame()) nframe_max =  Hdr().iFrame();
	t2 =Hdr().Epoch();
	icnt++;
	*phdr = Hdr();
	
      }
    }
    if( !feof(_fp) && ! SeekFrames(1) ){
      cerr <<"E VDIF_Frame::Read_1sec_hdr(): Failed SeekFrames(1)" << endl;
      break;
    }
    
  } // end while

  
  if( ( ( year ==0 && phdr->HD_Match() ) || 
	( year > 0 && phdr->HD_Match("Y", &yr) ))
      && t2 > t1 ){
    if(verbose_no >= _VERB_INFO )
      fprintf(stderr,"I Next first frame(%d) %s Detected.\n",
	      phdr->iFrame(), phdr->Epoch().Print_time(14).c_str());
    //      fseek(_fp,- framesize,SEEK_CUR);
    if(! jump_end )  SeekFrames(-1);
    icnt--; // redudce for next epoch.
    
  }else{
    if(EndofFile_Yes())
      cerr << " End of File.\n";
    else
      cerr << "E VDIF_Frame::Read_1sec_hdr(): header not found at the end." 
	   << endl;
  }
  // Set DAS information for latter use of Srate
  //  Set_Sparam(nch,qbit,srate );
  
  delete [] dummy;
  return icnt;
    
}

/******************************************/
const long VDIF_Frame::Read_every_sec_hdr(vector <VDIF_Hdr>& hdrarr,
					  const size_t nsec, const Read_t mode){
  if(_fp==0){
    cerr << "E Failed VDIF_Frame::Read_every_sec_hdr(): File is not opened yet\n";
    return -1;
  }

  long offset[2]={0,0}; /// First reading Header
  if(! VDIF_Hdr_find(_fp, offset[0], &_vdif_hdr)){
    if( ! EndofFile_Yes())   cerr << "Error: Failed to find VDIF-Header \n";
    return -1;
  }; 
  // The first header pos in the file.
  if( _offset_0 < 0) _offset_0 = offset[0];

  _hdr_tested = 0; ///< Set test-flag to 'OK'
  //  hdrarr.resize(0);
  long hdrsize = _vdif_hdr.Headsize();
  framesize = _vdif_hdr.Framesize();
  char *dummy = new char[hdrsize];
  VDIF_Hdr* phdr = (VDIF_Hdr*)dummy;

  if( !Find_FirstFrame() ){
    cerr <<"E VDIF_Frame::Read_every_sec_hdr(): Failed Find_FirstFrame()\n";
    return -1;
  }
  offset[1] = ftell(_fp);
  long frm_count = (offset[1] - offset[0])/framesize;  
  if(verbose_no >= _VERB_INFO)
    fprintf(stderr,"I VDIF_Frame::Read_every_sec_hdr(): First iframe0, %ld\n",frm_count);

  int sec_cnt=0;
  int nerr=10;
  //  nframe_sec =0;
  while( nsec-sec_cnt >0 && nerr > 0 ){
    offset[0] = offset[1];    
    if(fread( (void*)dummy, hdrsize, 1, _fp) < 1) break;
    fseek(_fp,framesize-hdrsize,SEEK_CUR);
    if( ! (phdr->HD_Match() && phdr->iFrame() ==0) ){
      nerr--; 
      if( !Find_FirstFrame() ){
	fprintf(stderr,"E VDIF_Frame::Read_every_sec_hdr(): Failed Find_FirstFrame() at %d sec.\n",
		sec_cnt);
	return -1;
      }
      offset[1] = ftell(_fp);
      frm_count = (offset[1] - offset[0])/framesize;  
      fprintf(stderr,"W VDIF_Frame::Read_every_sec_hdr(): %ld frame_sec\n", frm_count);
      continue;
    }
    hdrarr.push_back(*phdr);
    hdrarr[sec_cnt] = (*phdr);
    sec_cnt++;

    if(nframe_sec ==0){
      switch(phdr->EDV()){
      case EDV_NICT:
      case EDV_NRAO:
      case EDV_VLA:
      case EDV_NPDF:
	{
	  int pld_size = Framesize() - VDIF_Hdr_size;
	  nframe_sec = phdr->Speed()*4/pld_size;
	}
	break;
      case EDV_HAYSTACK:
      default:
	if(hdrarr.size() == 1){
	  if( !Find_FirstFrame() ){ // Net iframe=0
	    fprintf(stderr,"E VDIF_Frame::Read_every_sec_hdr(): Failed Find_FirstFrame() at %d sec.\n",
		    sec_cnt);
	    return -1;
	  }
	  fseek(_fp, -framesize, SEEK_CUR);
	  if(fread( (void*)dummy, hdrsize, 1, _fp) < 1) break;
	  fseek(_fp, framesize-hdrsize, SEEK_CUR);
	  nframe_sec = phdr->iFrame()+1;
	}
	break;
      } // Switch
    }
    
    if( !Find_FirstFrame() ){ // Net iframe=0
      fprintf(stderr,"E VDIF_Frame::Read_every_sec_hdr(): Failed Find_FirstFrame() at %d sec.\n",
	      sec_cnt);
      return -1;
    }
    
    offset[1] = ftell(_fp);    
    frm_count = (offset[1] - offset[0])/framesize;  
    if(verbose_no >= _VERB_INFO)
      fprintf(stderr,"I VDIF_Frame::Read_every_sec_hdr(): file address=%ld frame/sec\n", frm_count);

  }// While loop


  int nch = _vdif_hdr.Nch();
  int qbit = _vdif_hdr.Qbit();
  uint32_t srate(0);
  switch (hdrarr[0].EDV()){
  case EDV_NICT:
  case EDV_NRAO:
  case EDV_VLA:
  case EDV_NPDF:
    srate = _vdif_hdr.Srate();
    break;
  case EDV_HAYSTACK:
  default:
    // Set Srate for reference for non-NICT version
    if( frm_count>0){ 
      long datsize = Framesize() - Headsize();
      srate = nframe_sec * datsize * 8/qbit/nch;
      //      uint32_t srate_ksps = frm_sec0 * datsize * 8/hdrarr.size()/qbit/nch/1000;
      // vdif_hdr.Set_Srate(srate_ksps);
      // for(int i=0; i< (int)hdrarr.size(); i++) hdrarr[i].Set_Srate(srate_ksps);
    }
    break;
  }

  // Set DAS information for latter use of Srate
  Set_Sparam(nch,qbit,srate );
    
  delete [] dummy;
  return frm_count;
}

// const long VDIF_Frame::Read_every_sec_hdr(vector <VDIF_Hdr>& hdrarr,
// 					  const size_t nsec, const Read_t mode)
// {
//   if(_fp==0){
//     cerr << "E Failed VDIF_Frame::Test_Read(): File is not opened yet\n";
//     return -1;
//   }

//   long offset(0); /// First reading Header
//   if(! VDIF_Hdr_find(_fp, offset, &_vdif_hdr)){
//     cerr << "Error: Failed to find VDIF-Header \n";
//     return -1;
//   }; 

//   _hdr_tested = 0; ///< Set test-flag to 'OK'
//   //  hdrarr.resize(0);
//   framesize = _vdif_hdr.Framesize();
//   char *dummy = new char[framesize];
//   VDIF_Hdr* phdr = (VDIF_Hdr*)dummy;


//   // Search for the next 1sec
//   int sec_count(0);
//   long frm_sec_cnt=-1;  // Frame counter in every sec
//   //  long frm_sec0=0; // Estimate of Nframes/sec
//   //  long frm_intvl(0);
//   long frm_count(0);
//   nframe_sec =0;
//   while( nsec == 0 || (sec_count < (int)nsec) ){
//     if(!VDIF_Hdr_find(_fp, offset, phdr)) break;
//     if(fseek(_fp,framesize,SEEK_CUR) <0) break;
//     if(offset != 0) _hdr_tested++;
//     frm_count++;
//     frm_sec_cnt++;
//     if( phdr->iFrame() ==0 ){ // at initial integer sec.
//       if( (int)hdrarr.size() < sec_count+1)  hdrarr.push_back(*phdr);
//       else hdrarr[sec_count] = *phdr;
//       sec_count++;    
      
//       switch(phdr->EDV()){
//       case EDV_NICT:
//       case EDV_NRAO:
//       case EDV_VLA:
//       case EDV_NPDF:
// 	if(nframe_sec == 0){
// 	  int pld_size = Framesize() - VDIF_Hdr_size;
// 	  nframe_sec = phdr->Speed()*4/pld_size;
// 	}
// 	if(mode == QUICK) {
// 	  fseek(_fp,framesize*(nframe_sec - 1),SEEK_CUR); // Skip 1sec
// 	  frm_count += nframe_sec - 1;
// 	}
// 	break;
//       case EDV_HAYSTACK:
//       default:
// 	if(sec_count > 1 && nframe_sec == 0 )  nframe_sec = frm_sec_cnt;
// 	// if(nframe_sec == 0 || frm_sec0 != frm_sec_cnt){
// 	// nframe_sec = frm_sec0;
// 	if(nframe_sec > 0 && mode == QUICK) {
// 	  fseek(_fp,framesize*(nframe_sec - 1),SEEK_CUR); // Skip 1sec
// 	  frm_count += nframe_sec - 1;
// 	}
// 	break;
//       } // Switch
      
//       frm_sec_cnt=0;
//     };  // at initial integer sec 

//   } // while
//   //  if(frm_intvl >0) nframe_sec = frm_intvl;
//   //  nframe_sec = frm_count/sec_count;

//   int nch = _vdif_hdr.Nch();
//   int qbit = _vdif_hdr.Qbit();
//   uint32_t srate(0);
//   switch (hdrarr[0].EDV()){
//   case EDV_NICT:
//   case EDV_NRAO:
//   case EDV_VLA:
//   case EDV_NPDF:
//     srate = _vdif_hdr.Srate();
//     break;
//   case EDV_HAYSTACK:
//   default:
//     // Set Srate for reference for non-NICT version
//     if( frm_count>0){ 
//       long datsize = Framesize() - Headsize();
//       srate = nframe_sec * datsize * 8/qbit/nch;
//       //      uint32_t srate_ksps = frm_sec0 * datsize * 8/hdrarr.size()/qbit/nch/1000;
//       // vdif_hdr.Set_Srate(srate_ksps);
//       // for(int i=0; i< (int)hdrarr.size(); i++) hdrarr[i].Set_Srate(srate_ksps);
//     }
//     break;
//   }

//   // Set DAS information for latter use of Srate
//   Set_Sparam(nch,qbit,srate );
    
//   delete [] dummy;
//   return frm_count;
// }

const long VDIF_Frame::Read_every_hdr(vector <VDIF_Hdr>& hdrarr, const long nread ){
  if(_fp==0){
    cerr << "E Failed VDIF_Frame::Test_Read(): File is not opened yet\n";
    return -1;
  }

  long offset(0); /// First reading Header
  if(! VDIF_Hdr_find(_fp, offset, &_vdif_hdr)){
    if( ! EndofFile_Yes())   cerr << "Error: Failed to find VDIF-Header \n";
    return -1;
  }; 
  // The first header pos in the file.
  if( _offset_0 < 0) _offset_0=offset;

  _hdr_tested = 0; ///< Set test-flag to 'OK'
  hdrarr.resize(nread);
  framesize = _vdif_hdr.Framesize();
  char *dummy = new char[framesize];
  VDIF_Hdr* phdr = (VDIF_Hdr*)dummy;
  fseek(_fp,framesize,SEEK_CUR);
  if(nframe_sec ==0) Nframe_sec_estimate();
  
  //  hdrarr.push_back( _vdif_hdr);
  hdrarr[0]=_vdif_hdr;
  
  // Search for the next 1sec
  //  long frm_sec0=0;
  //  long frm_intvl(0);
  long frm_count(0);
  for(long i=1; i<nread && ! feof(_fp); i++){
    if(!VDIF_Hdr_find(_fp, offset, phdr)) break;
    if(fseek(_fp,framesize,SEEK_CUR) <0) break;
    if(offset != 0) _hdr_tested++;
    hdrarr[i]=(*phdr);
    // if( phdr->iFrame() == 0 ){
    //   nframe_sec = frm_count - frm_sec0 ;
    //   frm_sec0 = frm_count;
    // }
    frm_count++;
  } 


  int nch = _vdif_hdr.Nch();
  int qbit = _vdif_hdr.Qbit();
  uint32_t srate(0);
  switch (hdrarr[0].EDV()){
  case EDV_NICT:
  case EDV_NRAO:
  case EDV_VLA:
  case EDV_NPDF:
    srate = _vdif_hdr.Srate();
    break;
  case EDV_HAYSTACK:
  default:
    // Set Srate for reference for non-NICT version
    // if( frm_count>0){ 
        //   srate = frm_sec0 * datsize * 8/hdrarr.size()/qbit/nch;
    // }
    if( nframe_sec != 0){
      long datsize = Framesize() - Headsize();
      srate = nframe_sec * datsize/hdrarr.size()/qbit/nch * 8;
    }
    break;
  }

  // Set DAS information for latter use of Srate
  Set_Sparam(nch,qbit,srate );

  // // Set Srate for reference for non-NICT version
  // switch (hdrarr[0].EDV()){
  // case  EDV_NICT:
    
  //   && frm_count>0){ 
  //   long datsize = Framesize() - Headsize();
  //   int nch = hdrarr[0].Nch();
  //   int qbit = hdrarr[0].Qbit();
  //   uint32_t srate_ksps = nframe_sec * datsize * 8/qbit/nch/1000;
  //   _vdif_hdr.Set_Srate(srate_ksps);
  //   for(int i=0; i< (int)hdrarr.size(); i++) hdrarr[i].Set_Srate(srate_ksps);
  // }

  // // Set DAS information for latter use of Srate
  // Set_Sparam(vdif_hdr.Nch(), vdif_hdr.Qbit(), vdif_hdr.Srate() );

  delete [] dummy;
  return frm_count;
  
}

const long VDIF_Frame::Read_sec( vector <uint8_t>& rawbuf, const int sec){
  vector <VDIF_Hdr> hdrarr;
  if( Tested() < 1 ||  Read_every_sec_hdr(hdrarr, sec) <=0){
      cerr << "  E Failed to VIDF_Frame::Read_sec()" << endl;
      return -1;
  };

  long pktsize = hdrarr[0].Framesize();
  size_t bufsize = nframe_sec* pktsize * sec;
  if(rawbuf.size() != bufsize)  rawbuf.resize(bufsize);
  long cnt = fread( &rawbuf[0], bufsize, 1, _fp);

  long nframe = cnt*bufsize/pktsize;
  return nframe;
}

const ReturnFrameInfo VDIF_Frame::Read_frame( vector <VDIF_Hdr>& hdrarr, 
				   vector <uint8_t>& datbuf,
				   const int nframe)
{
  ReturnFrameInfo ret_frameinfo;
  
  if(feof(_fp)){
    ret_frameinfo.Set_END();
    return ret_frameinfo;
  }

  if( _workbuf.size() < 64) _workbuf.resize(64);
  // Set Pointers 
  VDIF_Hdr *phdr =  (VDIF_Hdr*)&_workbuf[0];
  size_t hdsize = phdr->Headsize();
  
  // Test reading frist buffer
  if(  _workbuf.size() < hdsize){
    _workbuf.resize(hdsize);
    phdr =  (VDIF_Hdr*)&_workbuf[0];
  }

  if( fread( &_workbuf[0], hdsize, 1, _fp) < 1){
    string msg;
    if(feof(_fp)){
      msg = "I VDIF_Frame::Read_frame():End of File.";
    }else{
      char buf[1024];
      msg = "E VDIF_Frame::Read_frame():Failed to read next hdr.";
      perror(msg.c_str()); 
      msg += strerror_r(errno,buf,1024);
    }
    ret_frameinfo.Set_NG(msg);
    return ret_frameinfo;
  }

  if( ! phdr->HD_Match() ){
    string msg = "E VDIF_Frame::Read_frame(): Header is missing.";
    if(verbose_no >= _VERB_WARN )  fprintf(stderr,"%s\n",msg.c_str());
    ret_frameinfo.Set_NG(msg);
    return ret_frameinfo;
  }

  fseek(_fp,-hdsize,SEEK_CUR); // Rewind back

  long framesize = phdr->Framesize(); 
  size_t bufsize = framesize * nframe;
  long datsize = framesize - hdsize;
  //  uint32_t nfrm_sec = Nframe_sec();
  _vdif_hdr = *phdr; // Set first data to the common hdr-data in the frame

  if(  _workbuf.size() < bufsize){
    _workbuf.resize(bufsize); 
    phdr =  (VDIF_Hdr*)&_workbuf[0];
  }
  
  
  size_t datbufsize = datsize * nframe;  
  if(  datbuf.size() != datbufsize)  datbuf.resize(datbufsize);
  if(  hdrarr.size() != (size_t)nframe)  hdrarr.resize(nframe);
  
  uint8_t *ptr = &_workbuf[0] + hdsize;
  uint8_t *ptrd0 = &datbuf[0];
  uint8_t *ptrd = ptrd0;
  
  int32_t iframe0 = _vdif_hdr.iFrame();
  int32_t iframe_prev = iframe0;
  uint frmcount=0; 
  //  uint32_t sec1 = _vdif_hdr.Sec_Ref();
  for(int i=0; i< (int)nframe; i++){
    if(feof(_fp)){
      ret_frameinfo.Set_END();
      return ret_frameinfo;
    }
    if(fread( (void*)phdr, framesize,1,_fp) < 1){
      // string msg = " VDIF_Frame::Read_frame():EOF found:";
      // char buf[1024];
      // msg += strerror_r(errno,buf,1024);
      // ret_frameinfo.Set_NG(msg);
      ret_frameinfo.Set_END();
      break;
    }

    if( i == 0 || iframe_prev > phdr->iFrame()){
      hdrarr[i] = *phdr;
      //      sec1 = phdr->Sec_Ref();
    }
    memcpy(ptrd, ptr, datsize);
    
    ptrd += datsize; 
    frmcount++;
    ret_frameinfo.Set_OK(frmcount);
    iframe_prev = phdr->iFrame();
    
    // else{
    //   fseek(_fp,-framesize,SEEK_CUR); // rewind by one framesize
    //   int jump = phdr->iFrame() - iframe_nxt;
    //   jump += jump < 0 ? nfrm_sec : 0;
    //   int jsec = phdr->Sec_Ref() - sec1;
    //   jump += jsec * nfrm_sec;
    //   ret_frameinfo.Set_Jump(frmcount, jump);
    //   fprintf(stderr,"W There was a jump of iFrame(): expected %d -> %d\n",
    // 	      iframe_nxt, phdr->iFrame());
    //   break;
    // }

  }

//   double dt = (double)datsize/4/DAS_Info().Speed();
//   Time_dat t1 = vdif_hdr.Epoch() + dt* vdif_hdr.iFrame(); 
//   Set_Epoch(t1);
  
  return ret_frameinfo;
}
  
const uint32_t VDIF_Frame::Nframe_sec(void){

  if(nframe_sec == 0 ){
    int32_t speed = _vdif_hdr.Speed();
    if( Hdr().EDV() == EDV_NICT || Hdr().EDV() == EDV_NRAO  || Hdr().EDV() == EDV_VLA){ 
      long datsize=_vdif_hdr.Framesize() - _vdif_hdr.Headsize();
      if( speed > 0) nframe_sec = speed*4/datsize;      
      else Nframe_sec_estimate();
    }else
      Nframe_sec_estimate();
  }

  return nframe_sec;
}
  
/** Set frame size by dividing pagesize(1sec) with a integer number */
void VDIF_Frame::Set_Nframe_sec(const uint32_t n){
    int32_t byte_sec = das_inf.Bytes_sec();
    int n1 =n;
    while( byte_sec % (8*n1) !=0 && byte_sec > 8*n1){
	n1++;
    }
    int32_t pld_size = byte_sec/n1;
    framesize = pld_size + VDIF_Hdr_size;
    _vdif_hdr.Set_FrameLen(framesize);
    if(verbose_no >= _VERB_INFO)
      fprintf(stderr, "I VDIF_Frame::Set_FrameDiv(%d) fixed Ndiv=%d, Framesize=%d\n",
	      n,n1,framesize);
    VLBI_Frame::nframe_sec = n1; // add 2014/5/22 miya
}

/** Find first iframe=0, if current data is so, no further pointer shift.*/
bool VDIF_Frame::Find_FirstFrame(void){

  if( _fp ==NULL){
    fprintf(stderr,"E VDIF_Frame::find_hdr(): vdif-file is not opened yet.\n");
    return  false;
  }
  
  long offset=0;
  if(! VDIF_Hdr_find(_fp, offset, &_vdif_hdr)){
    if(verbose_no >= _VERB_ERR)
      if( ! EndofFile_Yes())
	cerr << "E VDIF_Frame::Read_every_sec_hdr(): Failed to find VDIF-Header."  << endl;
    return false;
  }
  if( _offset_0 < 0) _offset_0=offset; // The first header pos in the file.
  
  if( Framesize() != _vdif_hdr.Framesize() ) Set_Head_par(_vdif_hdr);
  
  const int headsize=Headsize();
  unsigned char buf[headsize];
  VDIF_Hdr* phdr = (VDIF_Hdr*) buf;

  int32_t nframe_sec1=nframe_sec;
  if( nframe_sec ==0) nframe_sec1=Nframe_sec_estimate();
  if(nframe_sec1 < 0) return false;
  
  int one_skip=10000;
  int32_t prev_frame = _vdif_hdr.iFrame();
  uint32_t ntry = 1000000/one_skip;
  if( nframe_sec1 !=0){ // in case EUD_NICT, VLA, NRAO
    while( _vdif_hdr.iFrame() != 0 && ntry > 0){
      int32_t current_frame =  _vdif_hdr.iFrame();
      int nskip=nframe_sec - _vdif_hdr.iFrame();
      if(current_frame < prev_frame) nskip = -current_frame; 
      int nskip1 = nskip;
      if( abs(nskip1) > one_skip) { nskip1= nskip1 > 0 ? one_skip : - one_skip; } 
      //     fprintf(stderr,"I Skip=%d iframe=%d, ", nskip1, current_frame);

      SeekFrames(nskip1);
      fread( (void*)phdr, headsize, 1, _fp);
      if(! phdr->HD_Match()){
	if(verbose_no >= _VERB_WARN)
	  fprintf(stderr, "W VDIF_Frame::Read_every_sec_hdr():Sync not found at frame=%d\n",   current_frame);
	fseek(_fp, -headsize,SEEK_CUR);
	if(! VDIF_Hdr_find(_fp, offset, &_vdif_hdr)){
	  if( ! EndofFile_Yes())
	    cerr << "E VDIF_Frame::Read_every_sec_hdr(): Failed to find VDIF-Header." << endl;
	  return false;
	}

	if(verbose_no >= _VERB_WARN)
	fprintf(stderr,"W            Read_every_sec_hdr():Sync found for frame=%d\n", 
		_vdif_hdr.iFrame() );
	continue;
      } 
      prev_frame= current_frame;
      fseek(_fp, -headsize,SEEK_CUR);
      _vdif_hdr = *phdr;
      ntry --;
    }
  }else{ // In case Haystack, ALMA,NRAO
    long rate_fact[4]={1,2,4,8};
    long unit_frame =1;
    uint32_t Nmax=1000;
    int32_t pos_frame=_vdif_hdr.iFrame();
    ntry = Nmax;
// for(uint32_t i=0; i<Nmax &&  nframe_sec1 <= _vdif_hdr.iFrame() ; i++, ntry--){
    pos_frame = _vdif_hdr.iFrame();
    for(uint32_t i=0; i<Nmax && pos_frame <= _vdif_hdr.iFrame() ; i++, ntry--){
      pos_frame = _vdif_hdr.iFrame();
      int iord = i/4; unit_frame=1;
      for(int j =0; j<iord; j++) unit_frame *= 10; 
      nframe_sec1 = rate_fact[ i % 4 ] * unit_frame; 
      int  ipos = nframe_sec1 - pos_frame;
      if( ipos < 0 ) continue;
      SeekFrames( ipos ); 
      if(fread( (void*)&_vdif_hdr, headsize, 1, _fp) <1) throw "E Find_FirstFrame():Failed to read";
      fseek(_fp, -headsize,SEEK_CUR);
    }
    //    nframe_sec = nframe_sec1 - _vdif_hdr.iFrame();
    //    nframe_sec = nframe_sec1;
    SeekFrames(- _vdif_hdr.iFrame() );
    fread( (void*)&_vdif_hdr, headsize, 1, _fp);
    fseek(_fp, -headsize,SEEK_CUR);
  }

  if( ntry < 0){
    cerr << "E VDIF_Frame::Read_every_sec_hdr(): Failed to find First Frame(0).\n";
    return false;
  }
  if( _vdif_hdr.iFrame() ==0){
    //    if( _offset_0 < 0) _offset_0=ftell(_fp);
    return true;
  }else{
    return false;
  }
}

int32_t VDIF_Frame::Nframe_sec_estimate(void){

  long  headsize=Headsize();
  long framesize=Framesize();
  long datasize=framesize - headsize;
  // switch(_vdif_hdr.exud_NICT.ext_dat_ver){
  // case EDV_NICT:
  // case EDV_VLA :
  // case EDV_NRAO :
  // case EDV_NPDF:
  //   nframe_sec=_vdif_hdr.Speed()*4;
  //   nframe_sec /= datasize;
  //   return nframe_sec;
  //   break;
  // case EDV_NRAO4 :
  // case EDV_HAYSTACK:
  // default:
    {
      //      long pos_origin = ftell(_fp);
      
      unsigned char *buf = new unsigned char[framesize+128];
      VDIF_Hdr* phdr = (VDIF_Hdr*) buf;
      VDIF_Hdr hdr1= _vdif_hdr;

      int32_t jframe[2] =  { _vdif_hdr.iFrame() ,_vdif_hdr.iFrame() }; 
      //      float frame_slip_cnt=0;
      //      long rdsize=0;
      long pnt0= ftell(_fp);

      long tcnt=0;
      int ret=0;
      while( !feof(_fp) &&  jframe[0] <= jframe[1]*(1.0 + _frame_slip_margin) ){
	if(elecs_flag && tcnt == ELECS_NO){
	  ret= fread( (void*)buf, framesize+128, 1, _fp);
	  tcnt=0;
	}else
	  ret = fread( (void*)buf, framesize, 1, _fp);
	tcnt++;
	if(ret < 1 ){
	  perror("E VDIF_Frame::Nframe_sec_estimate()");
	  return 0;
	  //	  if( jframe[0] < 10) return 0;	  else break;
	}

	//	rdsize += framesize;
	if(! phdr->HD_Match())   continue;
	jframe[0]=jframe[1];
	jframe[1] =  phdr->iFrame() ;
      }
      int nch=phdr->Nch();
      int qbit=phdr->Qbit();
      double sr = 8.0* (jframe[0]+1)*datasize/nch/qbit*1.e-6; // Msps
      int isr = (int)ceil(sr);  // Msps
      float fs = 1.e6/8.0*isr*nch/datasize*qbit;
      nframe_sec = (uint32_t)fs;
      fseek(_fp,pnt0,SEEK_SET);
      _vdif_hdr.Set_Sparam(nch,qbit,isr*1000);
      das_inf.Set_Srate(isr*1000000);
      das_inf.Set_Nch(nch);
      das_inf.Set_Qbit(qbit);
    }
//     break;
// }
    
    return nframe_sec;
}

// int32_t VDIF_Frame::Nframe_sec_estimate(void){

//   long  headsize=Headsize();
//   long framesize=Framesize();
//   long datasize=framesize - headsize;
//   switch(_vdif_hdr.exud_NICT.ext_dat_ver){
//   case EDV_NICT:
//   case EDV_VLA :
//   case EDV_NRAO :
//   case EDV_NPDF:
//     nframe_sec=_vdif_hdr.Speed()*4;
//     nframe_sec /= datasize;
//     return nframe_sec;
//     break;
//   case EDV_NRAO4 :
//   case EDV_HAYSTACK:
//   default:
//     {
//       long rate_fact[4]={1,2,4,8};
//       long unit_frame =1;
//       long pos_origin = ftell(_fp);

//       uint32_t Nmax=1000;
//       int32_t pos_frame=0;
//       VDIF_Hdr hdr1= _vdif_hdr;
//       if( hdr1.iFrame()  ==0){
// 	if(fread( (void*)&hdr1, headsize, 1, _fp) <1) return 0;
// 	fseek(_fp, Framesize()-Headsize(), SEEK_CUR);
//       }
      
//       long nframe_sec1=0, nframe_sec2=0;
//       int ntry = Nmax;
//       pos_frame = hdr1.iFrame();
//  //for(uint32_t i=0; i<Nmax &&  nframe_sec1 <= hdr1.iFrame() ; i++, ntry--){
//       for(uint32_t i=0; i<Nmax &&  pos_frame <= hdr1.iFrame() ; i++, ntry--){
// 	pos_frame = hdr1.iFrame();
// 	int iord = i/4; unit_frame=1;
// 	for(int j =0; j<iord; j++) unit_frame *= 10; 
// 	nframe_sec2 = nframe_sec1;
// 	nframe_sec1 = rate_fact[ i % 4 ] * unit_frame; 
// 	int ipos = nframe_sec1 - pos_frame;
// 	if(ipos <0) continue;
// 	if( !SeekFrames( ipos ) ) return 0;; 
// 	if(fread( (void*)&hdr1, headsize, 1, _fp) <1) return 0;
// 	fseek(_fp, -headsize,SEEK_CUR);
//       }
//       //      nframe_sec = nframe_sec1 - hdr1.iFrame();
//       nframe_sec = nframe_sec2;
//       fseek(_fp,pos_origin,SEEK_SET);
     
//     }
//     return nframe_sec;
//     break;
//   }
// }

void VDIF_Frame::Set_edv(const EDV_t edv, const char* stname, const uint32_t Hz){
  _vdif_hdr.Set_edv(edv);
  if(edv == EDV_NICT){
    if(stname !=0 ) _vdif_hdr.exud_NICT.Set_stname(stname);
    if(Hz !=0 ) _vdif_hdr.exud_NICT.Set_Srate(Hz/1000);
  }
}
