
#ifndef __VDIF_HEADER__
#define __VDIF_HEADER__
/** VDIF Class Header 
 * @Version $Rev: 269 $
 * @Author M.Sekido
 * @Date  $Date: 2017-11-14 21:37:16 +0900 (火, 14 11月 2017) $
 * $Id: VDIF_Header.hpp 269 2017-11-14 12:37:16Z vlbi $ 
 */

//#include "VLBI_Header.hpp"
#include  <iostream>
#include  <sstream>
#include  <ctime>
#include  <cstring>
#include  <cstdio>
#include  <cstdlib>
#include  <string>
#include  <vector>
#include "time_lib.hpp"
#include "Mk5B_Header.hpp"
using namespace std;

extern "C" {
#include  <stdint.h>
#include  <error.h>
#include  <errno.h>
}

const int VDIF_Hdr_size = 32;
const int DEFAULT_VDIF_FRAMESIZE = 1456; 
const long MAX_FRAMESIZE = 100000000; 
const long MIN_FRAMESIZE = 500;

inline char AsciiChr(const char a){
  if(a<32 || a> 126) return '*';
  else return a;
}

//========== VDIF ===from Here  ====================

/// VDIF Sync patter of EDV1(NICT) Version
const uint32_t VDIF_NICT_SYNC = 0xACABFEED;
const uint32_t VDIF_HYST_SYNC = 0xABADDEED;
const uint32_t VDIF_ALMA_SYNC = 0x00A5EA50;

enum sample_rate_t {KSPS, MSPS};
// EDV Number designation see http://vlbi.org/vdif/
  //  const  uint8_t EDV_STD=0, EDV_NICT=0x01, EDV_ALMA=0x02, EDV_VLA=0x03, 
  // EDV_NRAO4=0x04, EDV_NRAO=0x0c, EDV_HAYSTACK=0xab, EDV_NPDF=128; 

enum EDV_t {EDV_STD=0, EDV_NICT=0x01, EDV_ALMA=0x02, EDV_VLA=0x03, 
	     EDV_NRAO4=0x04, EDV_NRAO=0x0c, EDV_HAYSTACK=0xab, EDV_NPDF=128}; 


struct VDIF_EUD_ALMA{
  unsigned sync:24;    ///< Sync pattern 'A5EA5X'
  uint8_t  ext_dat_ver;  ///< Extended User data version Number
  uint32_t pic_stat_word;        ///< 
  uint32_t pkt_ser_num_lsw;        ///< Sync pattern 'ACABFEED'
  uint32_t pkt_ser_num_msw;        ///< Sync pattern 'ACABFEED'
  VDIF_EUD_ALMA(void):ext_dat_ver(EDV_ALMA){;}
};

struct VDIF_EUD_Haystack{
  Mk5B_Hdr _mk5b_hdr;        ///< Mark5B Header
};
struct VDIF_EUD_Haystack2{
  uint32_t _arr4[4];
};


struct VDIF_EUD_NRAO4{
  unsigned unused:16;    ///< Unused 
  unsigned mask_length:8;   ///< Validity Mask length
  uint8_t ext_dat_ver;  ///< Extended User data version Number
  uint32_t sync;        ///< Sync pattern 'ACABFEED'
  uint32_t high_bmask;  ///< Higher bit Mask
  uint32_t low_bmask;   ///< Lower bit Mask
  VDIF_EUD_NRAO4(void):ext_dat_ver(EDV_NRAO4){;}
};

struct VDIF_EUD_VLA{
  unsigned srate:23;    ///< Sampling Rate (ksps/Msps)
  unsigned sr_unit:1;   ///< Sampling Rate unit (0:ksps,1:Msps)
  uint8_t ext_dat_ver;  ///< Extended User data version Number
  uint32_t sync;        ///< Sync pattern 'ACABFEED'
  uint32_t LowIF_freq;  ///< Low IF Frequency 
  char    misc[8];      ///< Misc Data
  VDIF_EUD_VLA(void):ext_dat_ver(EDV_VLA){;}
};

// struct VDIF_EUD_NPDF{
//   unsigned srate:23;    ///< Sampling Rate (ksps/Msps)
//   unsigned sr_unit:1;   ///< Sampling Rate unit (0:ksps,1:Msps)
//   uint8_t ext_dat_ver;  ///< Extended User data version Number
//   uint32_t sync;        ///< Sync pattern 'ACABFEED'
//   uint32_t LowIF_freq;  ///< Low IF Frequency 
//   unsigned _rflb_khz:24;
//   unsigned char buf[5];
// public:
//   VDIF_EUD_VLA(void):ext_dat_ver(EDV_VLA){;}
//   uint RFlb();
// };




/**NICT Pulsar Data format */
struct U_npdf{
  unsigned _rflb_khz:24;
  unsigned char buf[5];
};
/// VDIF Extended user data ver1(NICT)
struct VDIF_EUD_NICT{
  unsigned srate:23;    ///< Sampling Rate (ksps/Msps)
  unsigned sr_unit:1;   ///< Sampling Rate unit (0:ksps,1:Msps)
  uint8_t ext_dat_ver;  ///< Extended User data version Number
  uint32_t sync;        ///< Sync pattern 'ACABFEED'
  char    DAST_name[8]; ///< DAS/Station name 
public:
  /// Default Constructor of VDIF_EUD_nict1
  VDIF_EUD_NICT( uint8_t eudv=(uint8_t)EDV_NICT): srate(32000),sr_unit(KSPS),ext_dat_ver(eudv),  sync(VDIF_NICT_SYNC){  }

  /// Initialize the header
  void Init(const char* name="KASHIM34", const uint8_t eudv=(uint8_t)EDV_NICT);
  /// Set Sampling Rate (kHz/MHz)
  void Set_Srate(unsigned long sr, const sample_rate_t Msps_unit=KSPS){
    srate=sr; sr_unit=Msps_unit;}
  /// Set Station name
  void Set_stname(const char *name){strncpy(DAST_name,name,8);}
  /// Returns Station name in string 
  void Get_Stname(string& name)const;
  void Set_RFlb(const unsigned long khz){
    U_npdf u;
    u._rflb_khz = khz;  
    memcpy(DAST_name,&u,8); 
  }
};

inline  void VDIF_EUD_NICT::Init(const char* name, const uint8_t eudv){
  ext_dat_ver = eudv; // NICT
  switch(eudv){
  case EDV_NICT:
  case EDV_NRAO:
  case EDV_NRAO4:
  case EDV_VLA:
    sync = VDIF_NICT_SYNC;  break;
  case EDV_HAYSTACK:
    sync = VDIF_HYST_SYNC;  break;
  case EDV_ALMA:
    sync = VDIF_ALMA_SYNC;  break;
  case EDV_NPDF:
    break;
  }
  strncpy(DAST_name,name,8);
}

inline void VDIF_EUD_NICT::Get_Stname(string& name)const{
  char buf[10]; 
  strncpy(buf,DAST_name,8); buf[8]='\0';
  name = buf;
}

/// VDIF Extended user data ver128(NPDF)
struct VDIF_EUD_npdf:public VDIF_EUD_NICT { ///< NICT Pulsar Data format
public:
  /// Set RF Lower boubd Freq. [kHz]
  void Set_RFlb(const unsigned long khz){
    U_npdf u;
    u._rflb_khz = khz;  
    memcpy(DAST_name,&u,8); 
  }
  /// Returns RF Lower boubd Freq. [kHz]
  uint32_t RFlb(void)const{ U_npdf *u = (U_npdf*)DAST_name;  return (uint32_t)u->_rflb_khz;}
};

struct VDIF_Hdr{
  /** Print Information level **/
  enum Prnt_t {ALL=1, CONST=20, VAR=21, ONELINE=22,ALL_ONE=2,RAW=3 };
  enum Data_t {Real_Sample, Complex_Sample};
  /** Seconds from reference Epoch **/
  unsigned  Sec_from_Ref:30;
  /** Flag for legacy compatibility, 0:32 Bytes header, 1:16 Bytes header*/
  unsigned  Legacy_flag:1;
  /** Invalid flag 1: Invalid, 0: Valid */
  unsigned  Invalid_flag:1;
  /** Data Frame Number */
  unsigned  Frame_Number:24;
  /** Reference Epoch half year counter since 2000.0 */
  unsigned  Ref_Epoch_hy_counter:6;
  /** unused  */
  unsigned  unused:2;
  /** Data Frame length -1, unit: 8 bytes  */
  unsigned  Frame_Length:24;
  /** Log2(nch)*/
  unsigned  Lg_Nch:5;
  /** VDIF Version Number */
  unsigned  Version_No:3;
  /** Station ID*/
  unsigned char Station_ID[2];
  /** Thread ID*/
  unsigned  Thread_Id:10;
  /** Quantization Bit/sample -1 */
  unsigned  Qbit_1:5;
  /** Comlex data flag. 1:Complex, 0: Real data */
  unsigned  Complex_data_flag:1;
  /** Extended user data 1 */
  VDIF_EUD_NICT exud_NICT;
public:
  /// Constructor 
  //  enum Copy_Mode_t {Full_cpy,Param_cpy};
  VDIF_Hdr(const char* name="KASHIM34", const uint8_t eudv=EDV_NICT):
    Frame_Length(DEFAULT_VDIF_FRAMESIZE/8),exud_NICT(eudv){
    Version_No=1;Thread_Id=0;Qbit_1=1; 
    strncpy((char*)Station_ID,"kb",2);
    exud_NICT.Set_stname(name);
  }
  /** Copy Basic common property of VDIF data */
  //  VDIF_Hdr(const VDIF_Hdr& hdr, const Copy_Mode_t mode=Full_cpy);
  ~VDIF_Hdr(){;  }  ///< Destructor 
  const bool Valid() const { return ( Invalid_flag ? false : true);}
  /// Returns Total second from Reference Epoch 
  uint32_t Sec_Ref()const{   return  (uint32_t)Sec_from_Ref;    }
  int32_t  iFrame()const{return Frame_Number;} ///< Returns Frame Number 
  int Nch() const {return (1 << Lg_Nch); }  ///< Returns Nch
  int Qbit()const {return Qbit_1+1;}  ///< returns Qunatization bits
  /// Sampling rate (Hz)
  uint32_t Srate()const{return Srate_ksps()*1000;}
  //  time_t Ref_Epoch() const;  ///< Reference Epoch
  Time_dat Ref_Epoch() const;  ///< Reference Epoch
  /**! Printing Header data
   * \param fmt pattern 1,2,20,21
   */
  void Print_hd(FILE *fp=stdout, const Prnt_t fmt=ALL){
    fprintf(fp,"%s\n", Str_hd(fmt).c_str()); fflush(fp);
  };
  /**! Header data printing to string 
   * \param fmt pattern 1,2,20,21
   * \return string 
   */
  string Str_hd(const Prnt_t fmt=VDIF_Hdr::ALL);
  /// return framesize including header (bytes)
  int Framesize() const {long fs = (Frame_Length)*8; 
    return ( fs < VDIF_Hdr_size ? VDIF_Hdr_size : fs); }
  /**!
   * \return header size (bytes)
   */
  int Headsize()const {return (VDIF_Hdr_size);}
  /**!
   * \return Time_dat data of the epoch
   */
  Time_dat Epoch()const;
  /// Return integer MJD of the epoch
  uint32_t Epoch_imjd(){return (uint32_t)(floor(Epoch().Mjd()));}
  /// Returns Second of the day
  double Epoch_sod() {return Epoch().Sod();}  ///< return sec of the day (sec)  
  /// Return header type
  //  VLBI_Hdr_t Header_type()const {return VDIF_Header_Type;}
  bool HD_Match(const char* ptnstr=0, int *arr=0 ) const;
  //{return (exud_NICT.sync == VDIF_NICT_SYNC ? true: false);}
  /// Setting  Nch 
  void Set_Nch(const int nch);
  /// Setting Station ID
  void Set_Stnid(const char *id){strncpy((char*)Station_ID,(char*)id,2);}
  /// Setting Quantization bit
  void Set_Qbit(const int nbit){Qbit_1 = nbit-1;}
  /// Setting Sampling Freq (ksps);
  void Set_Srate(const uint32_t ksps);
  /// Return Sampling rate in ksps
  uint32_t Srate_ksps()const; 
  /// Setting Framesize (Bytes)
  void Set_FrameLen(const uint32_t fbyte){Frame_Length = fbyte/8;}
  /// Setting Sampling Parameters (nch,qbit,srate) a nd Framesize at a time
  int  Set_Sparam(const int nch, const int qbit, const uint32_t Hz);
  /// Setup Epoch parametes;
  void Set_Epoch(const Time_dat &t1);
  /// Set Initial Reference Epoch. 
  /// When t1 is given, nearst ealier Epoch is automatically set.
  Time_dat Set_Epoch0(const time_t t1);
  //  void Set_Epoch(const time_t t1, const time_t t0_ref);
  /// Set Frame number 
  void Set_iFrame(const uint32_t iframe){ Frame_Number = iframe;}
  /// Counting up Frame Number ;
  void iFrame_plus(){Frame_Number++;}
  /// Returns Total Data rate in word(4 bytes)/sec
  int32_t Speed()const { 
    if(HD_Match()) return (Srate()/32 * Nch() * Qbit()); else return -1;
  }
  /// Returns Data Type (Real/Complex)
  string Dtype(void) {string ret = (Complex_data_flag == 1 ? "Cmplx" :"Real"); 
    return ret;}
  /// Get Station ID
  const char* Get_Stnid(string & id) const{ 
    char buf[3]; buf[0]=AsciiChr(Station_ID[0]);buf[1]=AsciiChr(Station_ID[1]);
    buf[2]='\0'; id=buf; return id.c_str();}
  const string Stnid() const{ 
    char buf[5]; buf[0]=AsciiChr(Station_ID[0]);buf[1]=AsciiChr(Station_ID[1]);
    buf[2]='\0';    string ret=buf;
    return ret; }
  const string Stname() const{
    string str="N/A"; 
    if(exud_NICT.ext_dat_ver == EDV_NICT)  exud_NICT.Get_Stname(str);
    return str; }
  const string Version(void){
    char str[256];
    sprintf(str, "Ver.%d.%d", Version_No,exud_NICT.ext_dat_ver);
    string ss = str;
    return ss;
  }
  const int ThreadID(void) {  return  Thread_Id; }
  /// Returns Extended User Data Version 
  const EDV_t EDV() const {return (EDV_t)exud_NICT.ext_dat_ver;}
  /// Set EUD Version
  void Set_edv(const uint8_t edv);
  void Set_RFlb(const unsigned long khz);
  unsigned long RFlb(void);
  const Data_t Data_Type(){return (Complex_data_flag == 1 ? Complex_Sample : Real_Sample);}
  void Set_valid_flag(const bool flag){  Invalid_flag = flag ? 0 : 1;}
  void Param_Cpy(const VDIF_Hdr& hdr);
  bool operator==(const VDIF_Hdr& hdr1) const;
};

inline void VDIF_Hdr::Set_edv(const uint8_t edv){ 
  exud_NICT.ext_dat_ver = edv;  
  switch(edv){
  case EDV_NICT:
  case EDV_NRAO:
  case EDV_NRAO4:
  case EDV_VLA:
    exud_NICT.sync = VDIF_NICT_SYNC;  break;
  case EDV_HAYSTACK:
    exud_NICT.sync = VDIF_HYST_SYNC;  break;
  case EDV_ALMA:
    exud_NICT.sync = VDIF_ALMA_SYNC;  break;
  case EDV_NPDF:
    break;
  }
}

inline bool VDIF_Hdr::HD_Match(const char* ptnstr, int *arr) const{
  bool ret=false;
  if(ptnstr != 0){
    int slen=strlen(ptnstr);
    for(int i=0; i<slen; i++){
      switch (toupper(ptnstr[i])){
      case 'Y': 
	if(Epoch().Year() != arr[i]) return false; // check year
	break;
      default: break;
      }
    }
  }

  switch( exud_NICT.ext_dat_ver ){
  case EDV_NICT :
  case EDV_VLA :
  case EDV_NRAO4 :
    ret = (exud_NICT.sync == VDIF_NICT_SYNC);
    break;
  case EDV_HAYSTACK :
    {
      Mk5B_Hdr *hdr = (Mk5B_Hdr *)&exud_NICT;
      ret = (hdr->sync_word == VDIF_HYST_SYNC);
    }
    break;
  case EDV_ALMA :
    {
      VDIF_EUD_ALMA *eud = (VDIF_EUD_ALMA *)&exud_NICT;
      ret = (  (eud->sync & VDIF_ALMA_SYNC) == VDIF_ALMA_SYNC );
    }
    break;
  default: // Other case
    { 
      int good_reason=0;
      int nch=Nch();
      int qbit = Qbit();
      int bs=nch * qbit;
      Time_dat trng[2];
      trng[0].Set_time(2002,1,0,0,0);
      trng[1].Set_time(2030,1,0,0,0);
      if(Epoch() > trng[0] && Epoch() < trng[1] ) good_reason++;
      if(Ref_Epoch() > trng[0] && Epoch() < trng[1] ) good_reason++;
      if(Nch() >0 && Nch() < 1000 && Qbit() > 0 && Qbit() < 30) good_reason++;
      if( Framesize() > MIN_FRAMESIZE && Framesize() < MAX_FRAMESIZE )  good_reason++;
      if( good_reason > 3) ret=true;
      
      if(Epoch() < trng[0] || Epoch() > trng[1] ) ret = false; 
      else{
	if(nch <=0 || qbit <= 0 || bs == 0 || nch > 1000000) ret = false;
	else{
	  int frmsize = Framesize();
	  if(  frmsize <= 0  ) ret=false;
	  else{
	    if( (frmsize * 8) % bs != 0 ) ret=false;
	  }
	}
      }
    }
    break;
  }
  return ret;
}


/** mode=0: Copy only basic property */
inline void VDIF_Hdr::Param_Cpy(const VDIF_Hdr& hdr){
  Qbit_1 = hdr.Qbit_1;
  Lg_Nch = hdr.Lg_Nch;
  Complex_data_flag = hdr.Complex_data_flag;
  Legacy_flag = hdr.Legacy_flag;
  Invalid_flag = hdr.Invalid_flag;
  Sec_from_Ref=hdr.Sec_from_Ref;
  Ref_Epoch_hy_counter=hdr.Ref_Epoch_hy_counter;
  Frame_Length=hdr.Frame_Length;
}

inline  unsigned long VDIF_Hdr::RFlb(void){
  if(exud_NICT.ext_dat_ver == EDV_NPDF){
    VDIF_EUD_npdf *ptr = (VDIF_EUD_npdf *)&exud_NICT;
    return ptr->RFlb();
  }else{
    return 0;
  }

}

inline void VDIF_Hdr::Set_RFlb(const unsigned long khz){
  exud_NICT.Set_RFlb(khz);
  Set_edv(EDV_NPDF);
}

inline  void VDIF_Hdr::Set_Srate(const uint32_t ksps){
  if(ksps <1000000){
    exud_NICT.Set_Srate(ksps, KSPS);
  }else{
    exud_NICT.Set_Srate(ksps/1000,MSPS);
  }
}

inline  uint32_t VDIF_Hdr::Srate_ksps()const{
  switch(exud_NICT.ext_dat_ver){
  case EDV_NICT:
  case EDV_NRAO:
  case EDV_VLA:
    if(exud_NICT.sr_unit){ // Msps
      return exud_NICT.srate*1000;
    }else{ // ksps
      return exud_NICT.srate;
    }
    break;
  case EDV_HAYSTACK:
  default:
    return 0;
    break;
  }
  
}

inline  void VDIF_Hdr::Set_Nch(const int nch){
  int lch =0;
  int ich=1;
  while( (nch & ich) == 0){ ich <<= 1; lch++;}
  Lg_Nch = lch;
}

inline Time_dat VDIF_Hdr::Epoch()const{
  Time_dat t0;
  t0 =  Ref_Epoch();
  t0 += Sec_from_Ref;
  return t0;
}


enum VDIF_Ext_t {No_ext, NICT_ext };
/*!  Header Finder from buffer data
 *  \param buf data buffer pointer for input
 *  \param nbyte buffer size in byte for input.
 *  \param offset position of detected header for output
 *  \param hdr pointer of the header container for output
 ***/

const bool VDIF_Hdr_find(const uint8_t*  buf, const uint32_t nbyte,
			 long &offset, VDIF_Hdr* hdr);

/*!  Header Finder from a file.  
 *  \param fp  File pointer for input
 *  \param offset position of detected header for output
 *  \param hdr pointer of the header container for output
 *  File pointer is set to the top of header.
 ***/
const bool VDIF_Hdr_find(FILE *fp, long &offset, VDIF_Hdr* hdr);

// /// @class VDIF_Header is header class of VDIF
// class VDIF_Header: public VLBI_Header{
//   VDIF_Hdr* pt_head;
// public:
//   VDIF_Header():VLBI_Header(VDIF_Hdr_size),pt_head(0){;} /// < Default Constructor 
//   /// Header finder
//   /// @return  Number of bytes to find the header, or -1: for Error
//   int32_t Hd_find(const uint8_t*  buf, const uint32_t nbyte);
//   /// Print Header Information
// };

//========== VDIF ===to   Here  ====================

#endif
