

#ifndef __VIRDUMPER__
#define __VIRDUMPER__
/** VirDumper (Output) class library
 * @Version $Rev: 157 $
 * @Author M.Sekido
 * @Date  $Date: 2015-02-06 11:46:23 +0900 (釁E 06  2朁E2015) $
 * $Id: VirDumper.hpp 157 2015-02-06 02:46:23Z vlbi $ 
 */

#include <iostream>
#include "Thread.h"
#include <cstdio>
#include <string>
#include <cstdlib>
#include <cstring>
using namespace std;

#include "SUDPProto.hpp"
#include "VLBI_Frame.hpp"
#include "VTProto.hpp"
#include "VDIFProto.hpp"  //add miya 2017/12/19
#include "InVSI_mdl.hpp"
#include "Logsystem.hpp"

extern "C"{
  #include <time.h>
  #include <sys/time.h>
  #include <unistd.h>
}

//enum VirDumper_state_t {RUNNING, IDLE};
/** VirDumper
 * Function: Output data to Net/File. When multiple VirDumper instances are cleated and
 * used for output data streaming, Input model could be shared, but Protocol model 
 * instances must be independents at Set_IO(), because VLBI_Proto has socket and VLBI_Frame
 * pinters in the protocol should be independent for multile output.
 * $Author: vlbi $
 * $Revision: 157 $ $Date: 2015-02-06 11:46:23 +0900 (釁E 06  2朁E2015) $
 * $Id: VirDumper.hpp 157 2015-02-06 02:46:23Z vlbi $
 */

const int MAX_NOUT=16; //2014/12/9 add miya
const int LENGTH=86400; // add miya 2018/3/7
class VirDumper: public Thread{
public: 
  /*! DAS setting mode: 
   *  - IN2OUT   : Copying inDAS to outDAS
   *  - USROUTSET: outDAS  is set from ext(user). 
   *    outDAS must be subset of inDAS. 
   */
  enum DasMode_t {IN2OUT, USR_OUT_SET};
  /** Default Constructor  */
  VirDumper();
  /** Default Desructor  */
  virtual ~VirDumper(); 
  virtual void* thread_func()=0; ///< Thread function
  //  void Set_IO(FILE *out, InDev_Model *k){fp=out, input_mdl=k; }
  /// Setting IN(K5/VSI Device) and Out(protocol) 
  virtual int Set_IO(VLBI_Proto *v, InDev_Model *k, const int ith=0)=0; 
  //  void Set_IO(SUDPProto *s, K5_VSI *k){psudp =s; pk5vsi=k;}
  //int Set_IOarr(VLBI_Proto *v, InDev_Model *k, int ith); // add 2014/12/9 miya
  virtual int Writing(const time_t s_epoch=0)=0; ///< Main function for writing data 
  virtual int Sending()=0; ///< Main function to sending data 
  //  int Sending_sudp(const time_t s_epoch=0);
  int Set_Span(const int ts); ///< Set time length for dumping (sec)
  /// Set start time, Expecting format is %Y-%j-%H:%M:%S";
  void Set_Stime(string & st); 
  /// Set Statrt epoch
  void Set_Stime(const time_t time1){st_epoch=time1;}
  int Set_Reclength(const int length);  // add miya 2018/3/7
  virtual bool Ready()=0; ///< Check/Make ready
  void Stop(); ///< Stop dumping 
  /// Set DAS mode DasMode_t
  void SetDasMode(const DasMode_t m){dasMode = m;}
  void IntvlRegulator(); // Regulating the Tx1sec() interval to exct 1sec.
  int Set_Nout(const int ngrp); // add 2014/12/9 miya
protected: // comment 2015/3/6 miya
  //  VirDumper_state_t state; 
  //  FILE *fp;
///  VLBI_Proto* proto_mdl[MAX_NOUT];  ///< Output Protocol(including header format) used 2014/12/9 modify [16] miya comment 2015/3/6 miya
  VLBI_Frame* vlbi_frame; ///< Output VLBI frame pointer 
  //  SUDPProto *psudp;
  //  K5_VSI * pk5vsi;
  InDev_Model *input_mdl;
  int tspan;  // Time limit span (default 10) (sec), if negative, endless;
  time_t st_epoch;
  time_t end_epoch; // add miya 2018/3/7
  int rec_length; // add miya 2018/3/7
  DasMode_t dasMode;
  static int dth_number; ///< Dumber Thread number 
  struct timeval tm_reg; // Time memory for regulated interval 
  //  int Ndiv; // Numebr of data-division per second. add 2013/12/16 miya
  int NFileout;  //2014/12/9 add miya
};
/** Set in Input model and Output Protocol
 *  Then Input module started.
 */

inline  int VirDumper::Set_Nout(const int ngrp){
  NFileout = ngrp;
  return 0;
}

/// Reding time data in the format yyyy-ddd-hh:mm:ss 
inline void VirDumper::Set_Stime(string &st){
  struct tm tm1;
  string fmt1 = "%Y-%j-%H:%M:%S";
  strptime(st.c_str(), fmt1.c_str(), &tm1);
  setenv("TZ","GMT",1);tzset(); //add miya 2018/4/5
  st_epoch = mktime(&tm1);
}

inline int VirDumper::Set_Span(const int ts){
//   if(ts<0){ 
//     fprintf(stderr, "VirDumper::Set_span(): negative span(%d)", ts);
//     return -1; 
//   }else 
  tspan=ts;
  return 0;
}

inline int VirDumper::Set_Reclength(const int length){ // add miya 2018/3/8
  rec_length=length;
  return 0;
}


/** Destructor */
inline VirDumper::~VirDumper(){
  //  if(fp !=0){ fclose(fp); fp=0;}
  ///if( proto_mdl[0] != 0 ) proto_mdl[0] = 0; comment 2015/3/6 miya
  if( input_mdl != 0 ) input_mdl = 0;
  dth_number--;
}

inline void VirDumper::IntvlRegulator(){
  struct timeval tm1;
  gettimeofday(&tm1,0);
  double t_dif = (double)(tm1.tv_sec - tm_reg.tv_sec)
    + 1.e-6 * (tm1.tv_usec - tm_reg.tv_usec);
  
  ///if(t_dif <2.){
  if(t_dif <0.9){
    useconds_t ut = (useconds_t)((1.0 - t_dif)*1.e6);
  	usleep(ut);
//     struct timespec tmn, tmn2;
//     tmn.tv_sec =0;
//     tmn.tv_nsec = (ulong)((1.0 - t_dif)*1.e9);
//     nanosleep(&tmn,&tmn2);
    //    if(tmn2.tv_nsec >0)  nanosleep(&tmn2,0);
  }
  
  char buf[1024];
  struct timeval tm2;
  gettimeofday(&tm2,0);
  if( Logsystem::Test_Logsys(Logsystem::INFO,Logger::SMALL)
      && t_dif <2. ){
    sprintf(buf,"Cur(%lu,%lu) -Prev(%lu,%lu) slept(%.3lf sec) Now(%lu,%lu)\n",
	    tm1.tv_sec,tm1.tv_usec, tm_reg.tv_sec,tm_reg.tv_usec,(1. - t_dif),
	    tm2.tv_sec,tm2.tv_usec);
    string msg = "I VirDumper[";
    msg += Thread::StrID() + "]::IntvlRegulator(): " + buf; 
    Logsystem::info_logger()->Tstamp()  << msg;
    cerr      << msg;
  }

  tm_reg = tm2;

}

#endif
