/** 
 * @Version $Rev: 235 $
 * @Author M.Sekido
 * @Date  $Date: 2017-09-18 23:14:38 +0900 (月, 18  9月 2017) $
 * $Id: VsiData.cpp 235 2017-09-18 14:14:38Z vlbi $ 
 */

#include "VsiData.hpp"
#include <cstring>

int VsiData::Make_Table(const BitEnc_t enc_type){
  if(_dasInfo ==0){
    cerr << "Error @VsiData::Make_Table(): dasInfo is empty.\n";
    return -1;
  }
  try{
    switch(_dasInfo->Qbit()){
    case 1: 	MkTbl1bit();	break;
    case 2: 	MkTbl2bit(enc_type);	break;
    case 4: 	MkTbl4bit();	break;
    case 8: 	MkTbl8bit();	break;
    default:	throw "Impossible Qbit"; break;
    }
  }

  catch(const char* msg){
    cerr << "Error @VsiData::Make_Table(): " << msg << endl;
    return -1;
  }
  return 0;
}


const int VsiData::MkTbl1bit(){
  if(_dasInfo ==0){  
    cerr << "Error @VsiData::MkTbl4_1(): dasInfo is empty.\n";
    return -1;  
  }
  for(int i=0; i<256; i++){
    bzero(_table[i], MAX_NBIT*sizeof(int8_t));
    bzero(_ftable[i], MAX_NBIT*sizeof(float));
  }
  int8_t *ptr;
  float  *ptrf;
  for(int i=0; i<256; i++){
    ptr =  &_table[i][0];
    if(i & 0x01)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; ptr++;
    if(i & 0x02)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; ptr++;
    if(i & 0x04)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; ptr++;
    if(i & 0x08)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; ptr++;
    if(i & 0x10)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; ptr++;
    if(i & 0x20)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; ptr++;
    if(i & 0x40)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; ptr++;
    if(i & 0x80)  *ptr=level_1bit[1]; else *ptr=level_1bit[0]; 

    ptrf = &_ftable[i][0]; // For float 
    if(i & 0x01)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; ptrf++;
    if(i & 0x02)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; ptrf++;
    if(i & 0x04)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; ptrf++;
    if(i & 0x08)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; ptrf++;
    if(i & 0x10)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; ptrf++;
    if(i & 0x20)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; ptrf++;
    if(i & 0x40)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; ptrf++;
    if(i & 0x80)  *ptrf=(float)level_1bit[1]; else *ptrf=(float)level_1bit[0]; 
  }
  return 0;
}

const int VsiData::MkTbl2bit(const BitEnc_t enc_type){
  if(_dasInfo ==0){  
    cerr << "Error @VsiData::MkTbl4_1(): dasInfo is empty.\n";
    return -1;  
  }
  for(int i=0; i<256; i++){
    bzero(_table[i], MAX_NBIT*sizeof(uint8_t));
    bzero(_ftable[i], MAX_NBIT*sizeof(float));
    bzero(_fctable[i], MAX_NBIT*sizeof(FComplex));
  }

  switch(enc_type){
  case ADS3000_2bit:
    for(int i=0; i<256; i++){
      for(int j=0; j<4; j++){
	int v = (i & (0x03 << (j*2))) >> (j*2);
	_table[i][j]  =        level_ADS3000_2bit[v];	
	_ftable[i][j] = (float)level_ADS3000_2bit[v];	
      };
    }
    break;
  case SignMag_2bit:
    for(int i=0; i<256; i++){
      for(int j=0; j<4; j++){
	int v = (i & (0x03 << (j*2))) >> (j*2);
	_table[i][j]  =        level_SignMag_2bit[v];	
	_ftable[i][j] = (float)level_SignMag_2bit[v];	
      };
    }
    break;
  case RevSgnMag_2bit:
    for(int i=0; i<256; i++){
      for(int j=0; j<4; j++){
	int v = (i & (0x03 << (j*2))) >> (j*2);
	_table[i][j]  =        level_RevSgnMag_2bit[v];	
	_ftable[i][j] = (float)level_RevSgnMag_2bit[v];	
      };
    }
    break;
  case MS_2bit_1bit: // MSB side is taken for 1bit
    for(int i=0; i<256; i++){
      for(int j=0; j<4; j++){
	int v = (i & (0x02 << (j*2))) >> (j*2+1);
	_table[i][j]  =        level_1bit[v];	
	_ftable[i][j] = (float)level_1bit[v];	
      };
    }
    break;
  case LS_2bit_1bit: // LSB side is taken for 1bit
    for(int i=0; i<256; i++){
      for(int j=0; j<4; j++){
	int v = (i & (0x01 << (j*2))) >> (j*2);
	_table[i][j]  =        level_1bit[v];	
	_ftable[i][j] = (float)level_1bit[v];	
      };
    }
    break;
  case IQ_2bit: // IQ complex data with 2 bit
    for(int i=0; i<256; i++){
      for(int j=0; j<2; j++){
	int v1 = (i & (0x03 << (j*4))) >> (j*4);
	int v2 = (i & (0x03 << (j*4+2))) >> (j*4+2);
	_fctable[i][j] = FComplex((float)level_2bit[v1],(float)level_2bit[v2]);	
      };
    }
    _s_type = IQ_Sample;
    break;
  case IQ_1bit: // IQ complex data with 1 bit
    for(int i=0; i<256; i++){
      for(int j=0; j<4; j++){
	int v1 = (i & (0x01 << (j*2))) >> (j*2);
	int v2 = (i & (0x01 << (j*2+1))) >> (j*2+1);
	_fctable[i][j] = FComplex((float)level_1bit[v1],(float)level_1bit[v2]);	
      };
    }
    _s_type = IQ_Sample;
    break;
  case ADS2000_2bit:// This is natural and also be adopted by Makr5B and other VLBI terminal
  default:
    for(int i=0; i<256; i++){
      for(int j=0; j<4; j++){
	int v = (i & (0x03 << (j*2))) >> (j*2);
	_table[i][j]  =        level_2bit[v];	
	_ftable[i][j] = (float)level_2bit[v];	
      };
    }
    break;
  }

  return 0;
}


const int VsiData::MkTbl4bit(){
  if(_dasInfo ==0){  
    cerr << "Error @VsiData::MkTbl4_1(): dasInfo is empty.\n";
    return -1;  
  }
  for(int i=0; i<256; i++){
    bzero(_table[i], MAX_NBIT*sizeof(uint8_t));
    bzero(_ftable[i], MAX_NBIT*sizeof(float));
  }
  for(int i=0; i<256; i++){
    _table[i][0] = (i & 0x0f) - 8;
    _table[i][1] = ((i & 0xf0) >> 4) - 8;
    _ftable[i][0] = (float)_table[i][0];
    _ftable[i][1] = (float)_table[i][1];
  }
  return 0;
}

const int  VsiData::MkTbl8bit(){
  if(_dasInfo ==0){  
    cerr << "Error @VsiData::MkTbl4_1(): dasInfo is empty.\n";
    return -1;  
  }
  for(int i=0; i<256; i++){
    bzero(_table[i], MAX_NBIT*sizeof(uint8_t));
    bzero(_ftable[i], MAX_NBIT*sizeof(float));
  }
  for(int i=0; i<256; i++){ 
    _table[i][0] = i -128; 
    _ftable[i][0] = (float)_table[i][0]; 
  }
  return 0;
}

const int32_t VsiData::Read1Sec(FILE* fp){
  int32_t ByteSec = _dasInfo->Speed()*4;
  if( (int32_t)_buf.size() < ByteSec) _buf.resize(ByteSec);
  
  if( fread(&_buf[0],ByteSec,1,fp) < 1){
    cerr << "Error in Read1Sec()\n";
    return -1;
  }
  return ByteSec;
}

/// datbuf: Binary Data buffer, ndat: Data Numner, 
/// tgt_type: 0: Float, 1:Integer
const long VsiData::Interpret(vector <uint8_t>& datbuf, 
			      const long ndat, const int tgt_type){
#ifdef _DEBUG_
  cerr << "Interpretation...";
#endif
  if(_dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return -1;  
  }

  int nch= _dasInfo->Nch();
  int bit = _dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  if(_s_type == IQ_Sample){ // IQ data
    dunit*=2; 
    ch_byte /= 2; 
  }
  int unit_byte(1);
  int byte_unit (0);
  char msg[256];
  const size_t Npt_def=2048000;
  // size_t npt = ndat == 0 ? _dasInfo->Srate() : ndat;
  size_t npt = ndat == 0 ? Npt_def : ndat;
  if( nch > ch_byte){  // 1 unit needs more than 1 Byte
    if( dunit % 8 != 0){
      sprintf(msg,  "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit",  nch, bit);   
      cerr << "Err: " << msg << endl;      return -1;
    }
    byte_unit = dunit / 8;
  }else{  // 1 byte include no less than 1 unit
    if( 8 % dunit != 0 ){
      sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit",  nch, bit);   
      cerr << "Err: " << msg << endl;      return -1;
    }else{
      unit_byte = 8 / dunit;
    }
  }

  if(npt % unit_byte !=0){
    uint32_t ntmp = unit_byte;
    while(npt > ntmp ) ntmp += unit_byte;
    npt = ntmp;
  }

  if( _s_type == Real_Sample){  
    for( int i=0; i<nch; i++) {
      _valarr[i].resize(npt);
      _fvalarr[i].resize(npt);
    }
  }else{
    //    for( int i=0; i<nch; i++) { _fcarr[i].resize(npt); }
    for( int i=0; i<nch; i++) { _fcarr[i] = new FComplex[npt]; }
  }


  // vector <FComplex>::iterator cptr[nch];
  // vector <float>::iterator dptr[nch];
  // vector <int8_t>::iterator dptr2[nch];

  // for(int i=0; i<nch; i++){
  //   dptr[i] = _fvalarr[i].begin();  
  //   dptr2[i] = _valarr[i].begin();  
  // }

  uint32_t cnt=0;
  if(_s_type == Real_Sample){
    if( byte_unit == 0){ // 1 byte contains multiple units. //unit takes multiple bytes 
      if(tgt_type ==0) // Convertion to  float data
	cnt = Conv_data_f_unitbyte(datbuf,nch,unit_byte,npt);
      else   // Convertion to  Integer data
	cnt = Conv_data_d_unitbyte(datbuf,nch,unit_byte,npt);
      // vector <uint8_t>::iterator sptr = datbuf.begin();  
      // while(sptr < datbuf.end() && dptr[0] < _fvalarr[0].end() && cnt < npt){
      // 	for( int i=0; i< unit_byte; i++){
      // 	  for(int j=0; j<nch; j++){
      // 	    *dptr[j] = _ftable[*sptr][j+i*nch]; 
      // 	    *dptr2[j] = _table[*sptr][j+i*nch]; 
      // 	    dptr[j]++; dptr2[j]++;
      // 	  }
      // 	  cnt++;
      // 	}
      // 	sptr++;
      // } 
    }else{ // 1 unit takes multiple bytes
      //      Conv_data <float> C2(datbuf, nch, _fvalarr, _ftable, byte_unit, ch_byte, npt);
      if(tgt_type ==0) // Convertion to  float data
	cnt = Conv_data_f_byteunit(datbuf,  nch,  byte_unit,  ch_byte, npt); 
      else  // Convertion to  Integer data
	cnt = Conv_data_d_byteunit(datbuf,  nch,  byte_unit,  ch_byte, npt); 
      //   vector <uint8_t>::iterator sptr = datbuf.begin();  
      //   while(sptr < datbuf.end() && dptr[0] < _fvalarr[0].end() && cnt < npt){
      // 	int  ich=0;
      // 	for(int i=0; i<byte_unit; i++){ // bytes/unit
      // 	  for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
      // 	    *dptr[ich] = _ftable[*sptr][jch];  
      // 	    *dptr2[ich] = _table[*sptr][jch];  
      // 	    dptr[ich]++; 	  dptr2[ich]++; 
      // 	  }
      // 	  sptr++;
      // 	}
      // 	cnt++;
      //   }
      // } 
    }
  }else{ // Complex data
    if( byte_unit == 0){ // 1 byte contains multiple units. //unit takes multiple bytes 
      cnt  = Conv_data_c_unitbyte(datbuf,nch,unit_byte,npt);
      //      Conv_data_s <FComplex> Cs(datbuf, nch, _fcarr, _fctable, unit_byte, npt);
    }else{
      cnt = Conv_data_c_byteunit(datbuf,  nch,  byte_unit,  ch_byte, npt); 
      //   Conv_data <FComplex> C2(datbuf, nch, _fcarr, _fctable, byte_unit, ch_byte, npt);
    }
  }
#ifdef _DEBUG_
  cerr << "end" << endl;
#endif
  return cnt;
}


const long VsiData::Interpret( const long ndat){
#ifdef _DEBUG_
  cerr << "Interpretation...";
#endif
  if(_dasInfo ==0){  
    cerr << "Error @VsiData::Val():dasInfo is empty.\n";
    return -1;  
  }
  
  int nch= _dasInfo->Nch();
  int bit = _dasInfo->Qbit();
  int dunit = nch * bit;
  int ch_byte = 8/bit;
  int unit_byte(0);
  int byte_unit (0);
  char msg[256];
  try{
    if( nch > ch_byte){  // 1 unit takes more than 1 ch
      if( dunit % 8 != 0){
	sprintf(msg, 
		"Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }
      byte_unit = dunit / 8;
    }else{  // 1 byte include no less than 1 channel
      if( 8 % dunit != 0 ){
	sprintf(msg, "Nch(%d)*Qbit(%d) must be multiple of 8 bit or less than 8bit", nch, bit);
	throw msg;
      }else{
	unit_byte = 8 / dunit;
      }
    }
  } 
  catch(const char* msg1){
    cerr << "Error in VsiData::Val():" << msg1 << endl;
    return -1;
  }
  
  size_t npt = ndat;
  if(ndat == 0)  npt = _dasInfo->Srate();
  if(npt % unit_byte !=0){
    uint32_t ntmp = unit_byte;
    while(npt > ntmp ) ntmp += unit_byte;
    npt = ntmp;
  }

  for( int i=0; i<nch; i++) 
    if( _valarr[i].size() < npt) _valarr[i].resize(npt);
  vector <uint8_t>::iterator sptr = _buf.begin();  
  vector <float>::iterator dptr[nch];
  vector <int8_t>::iterator dptr2[nch];
  for(int i=0; i<nch; i++){
    dptr[i]  = _fvalarr[i].begin();  
    dptr2[i] = _valarr[i].begin();  
  }

  uint32_t cnt=0;
  
  if( byte_unit == 0){ // 1 byte contains multiple units
    while(sptr < _buf.end() && dptr2[0] < _valarr[0].end() && cnt < npt){
      for( int i=0; i< unit_byte; i++){
	for(int j=0; j<nch; j++){
	  *dptr2[j] = _table[*sptr][j+i*nch]; 
	  *dptr[j] = _ftable[*sptr][j+i*nch]; 
	  dptr[j]++;	  dptr2[j]++;
	}
	cnt++;
      }
      sptr++;
    } 
  }else{ // 1 unit takes multiple bytes
    while(sptr < _buf.end() && dptr2[0] < _valarr[0].end() && cnt < npt){
      int  ich=0;
      for(int i=0; i<byte_unit; i++){ // bytes/unit
	for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
	  *dptr2[ich] = _table[*sptr][jch];  
	  *dptr[ich] = _ftable[*sptr][jch];  
	  dptr[ich]++;	  dptr2[ich]++;
	}
	sptr++;
      }
      cnt++;
    }
  } 
#ifdef _DEBUG_
  cerr << "end" << endl;  
#endif
  return cnt;
}


/** Convert Integer data for multiple Units in a bytes */
const int VsiData::Conv_data_d_unitbyte(vector <uint8_t>& datbuf, const int nch, 
					const int unit_byte, const  size_t npt){
  vector <uint8_t>::iterator sptr = datbuf.begin();  
  vector <int8_t>::iterator ptr[MAX_NCH];
  for(int i=0; i<nch; i++){ ptr[i] = _valarr[i].begin();   }

  uint32_t cnt=0;
  while(sptr < datbuf.end() && ptr[0] < _valarr[0].end() && cnt < npt){
    for( int i=0; i< unit_byte; i++){
      for(int j=0; j<nch; j++){
	*ptr[j] = _table[*sptr][j+i*nch]; 
	ptr[j]++; 
	//	    dptr2[j]++;
      }
      cnt++;
    }
    sptr++;
  } 
  return cnt;
}

/** Convert Float data for multiple Units in a bytes */
const int VsiData::Conv_data_f_unitbyte(vector <uint8_t>& datbuf, const int nch, 
					const int unit_byte, const  size_t npt){
  vector <uint8_t>::iterator sptr = datbuf.begin();  
  vector <float>::iterator ptr[MAX_NCH];
  for(int i=0; i<nch; i++){ ptr[i] = _fvalarr[i].begin();   }

  uint32_t cnt=0;
  while(sptr < datbuf.end() && ptr[0] < _fvalarr[0].end() && cnt < npt){
    for( int i=0; i< unit_byte; i++){
      for(int j=0; j<nch; j++){
	*ptr[j] = _ftable[*sptr][j+i*nch]; 
	ptr[j]++; 
      }
      cnt++;
    }
    sptr++;
  } 
  return cnt;
}

/** Convert FComplex data for multiple Units in a bytes */
const int VsiData::Conv_data_c_unitbyte(vector <uint8_t>& datbuf, const int nch, 
					const int unit_byte, const  size_t npt){
  vector <uint8_t>::iterator sptr = datbuf.begin();  
  FComplex* ptr[MAX_NCH];
  for(int i=0; i<nch; i++){ ptr[i] = _fcarr[i];   }

  uint32_t cnt=0;
  while(sptr < datbuf.end() &&  cnt < npt){
    for( int i=0; i< unit_byte; i++){
      for(int j=0; j<nch; j++){
	*ptr[j] = _fctable[*sptr][j+i*nch]; 
	ptr[j]++; 
      }
      cnt++;
    }
    sptr++;
  } 
  return cnt;
}

// const int VsiData::Conv_data_c_unitbyte(vector <uint8_t>& datbuf, const int nch, 
// 					const int unit_byte, const  size_t npt){
//   vector <uint8_t>::iterator sptr = datbuf.begin();  
//   vector <FComplex>::iterator ptr[MAX_NCH];
//   for(int i=0; i<nch; i++){ ptr[i] = _fcarr[i].begin();   }

//   uint32_t cnt=0;
//   while(sptr < datbuf.end() && ptr[0] < _fcarr[0].end() && cnt < npt){
//     for( int i=0; i< unit_byte; i++){
//       for(int j=0; j<nch; j++){
// 	*ptr[j] = _fctable[*sptr][j+i*nch]; 
// 	ptr[j]++; 
//       }
//       cnt++;
//     }
//     sptr++;
//   } 
//   return cnt;
// }


const int VsiData::Conv_data_f_byteunit(vector <uint8_t>& datbuf, const int nch, 
		    const int byte_unit, const int ch_byte, const  size_t npt ){
  
  size_t cnt=0;
  vector <uint8_t>::iterator sptr = datbuf.begin();  
  vector <float>::iterator ptr[MAX_NCH];
  for(int i=0; i<nch; i++){ ptr[i] = _fvalarr[i].begin();   }

  while(sptr < datbuf.end() && ptr[0] < _fvalarr[0].end() && cnt < npt){
    int  ich=0;
    for(int i=0; i<byte_unit; i++){ // bytes/unit
      for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
	*ptr[ich] = _ftable[*sptr][jch];  
	ptr[ich]++;   
      }
      sptr++;
    }
    cnt++;
  }
  return cnt;
}

const int VsiData::Conv_data_d_byteunit(vector <uint8_t>& datbuf, const int nch, 
		    const int byte_unit, const int ch_byte, const  size_t npt ){
  
  size_t cnt=0;
  vector <uint8_t>::iterator sptr = datbuf.begin();  
  vector <int8_t>::iterator ptr[MAX_NCH];
  for(int i=0; i<nch; i++){ ptr[i] = _valarr[i].begin();   }

  while(sptr < datbuf.end() && ptr[0] < _valarr[0].end() && cnt < npt){
    int  ich=0;
    for(int i=0; i<byte_unit; i++){ // bytes/unit
      for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
	*ptr[ich] = _table[*sptr][jch];  
	ptr[ich]++;   
      }
      sptr++;
    }
    cnt++;
  }
  return cnt;
}

const int VsiData::Conv_data_c_byteunit(vector <uint8_t>& datbuf, const int nch, 
		    const int byte_unit, const int ch_byte, const  size_t npt ){
  
  size_t cnt=0;
  vector <uint8_t>::iterator sptr = datbuf.begin();  
  FComplex* ptr[MAX_NCH];
  for(int i=0; i<nch; i++){ ptr[i] = _fcarr[i];   }

  while(sptr < datbuf.end() && cnt < npt){
    int  ich=0;
    for(int i=0; i<byte_unit; i++){ // bytes/unit
      for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
	*ptr[ich] = _fctable[*sptr][jch];  
	ptr[ich]++;   
      }
      sptr++;
    }
    cnt++;
  }

  return cnt;
}

// const int VsiData::Conv_data_c_byteunit(vector <uint8_t>& datbuf, const int nch, 
// 		    const int byte_unit, const int ch_byte, const  size_t npt ){
  
//   size_t cnt=0;
//   vector <uint8_t>::iterator sptr = datbuf.begin();  
//   vector <FComplex>::iterator ptr[MAX_NCH];
//   for(int i=0; i<nch; i++){ ptr[i] = _fcarr[i].begin();   }

//   while(sptr < datbuf.end() && ptr[0] < _fcarr[0].end() && cnt < npt){
//     int  ich=0;
//     for(int i=0; i<byte_unit; i++){ // bytes/unit
//       for(int jch=0; jch< ch_byte; jch++,ich++){ // ch/byte
// 	*ptr[ich] = _fctable[*sptr][jch];  
// 	ptr[ich]++;   
//       }
//       sptr++;
//     }
//     cnt++;
//   }

//   return cnt;
// }
