/** 
 * @Version $Rev: 56 $
 * @Author M.Sekido
 * @Date  $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * $Id: crc.cpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

#include <cstdio>
#include <cstdlib>

#include "crc.hpp"

// crc subroutine - converts input into a 16 bit CRC code
const uint16_t CRC_lba::Cmp_CRC(const uint8_t *idata, const int nbyte) {
    
  int len = nbyte*8;

  uint16_t istate = 0;
  int idbit;
  int q, ich, icb;

  for (idbit = 1; idbit <= len; idbit++) {
    q = istate & 1;
    ich = (idbit - 1) / 8;
    icb = 7 - (idbit - 1) % 8;
    if ((((idata[ich] >> icb) & 1) ^ q)  ==  0)
      istate &= -2;
    else {
      istate ^=  mask;
      istate |= 1;
    }
    istate = (istate >> 1) | (istate & 1) << (cycl -1);
  } 
  return (istate);
}

const uint16_t CRC_mdl::reversebits16 (const uint16_t s) {
  unsigned char *c1, *c2;
  unsigned short ss;
  
  c1 = (unsigned char*)&s;
  c2 = (unsigned char*)&ss;
  
  // Flip bytes and bits
  c2[0] = bytetable[c1[1]]; 
  c2[1] = bytetable[c1[0]];
  return ss;
}

void CRC_mdl::init_bitreversal () {
  for (int i=0; i<256; i++) {
    bytetable[i] = ((i&0x80)>>7
                    |(i&0x40)>>5
                    |(i&0x20)>>3
                    |(i&0x10)>>1
                    |(i&0x08)<<1
                    |(i&0x04)<<3
                    |(i&0x02)<<5
                    |(i&0x01)<<7);
  }
}


const uint16_t CRC_k5vsi::Cmp_CRC(const uint8_t *dat, const int nbytes){
  /**
     crc Polynomial will be 
   CRC-16 : 0x1805  (default)
   CRC-CCITT: 0x11021
  **/
 
  uint32_t  work=0x00;
  uint32_t key = CRC_key();
  printf("crc_key= %X ", key);
  key <<= 8;
  printf("crc_key= %X \n", key);
  for(int i=0; i<nbytes; i++){
    work |= (0xff & dat[i]);
    int cnt=8;
    while(cnt-- ){
      work <<=1;
      if(work & 0x1000000) 
	work ^= key;
    }
  }
  

  for(int i=0; i<2; i++){
    int cnt=8;
    while(cnt-- ){
      work <<=1;
      if(work & 0x1000000) 
	work ^= key;
    }
  }
  work >>=8;
  return (work & 0xffff);
}


const uint16_t CRC_jive::Cmp_CRC(const uint8_t *dat, const int nbytes){
  Reset_crc();
  for(int i=0; i<nbytes; i++)  
    put_crc_byte(dat[i]);
  return (crc_register);
}

void CRC_jive::put_crc_byte(const uint8_t byte) {
  unsigned char top = (crc_register >> 8) ^ byte; // get the top byte of register
  crc_register = (crc_register << 8) ^ crc_table [top];
}

// void CRC_jive::put_crc_byte(const uint8_t byte) {
//   unsigned char top = crc_register >> 8; // get the top byte of register
//   crc_register = ((crc_register << 8) + byte) ^ crc_table [top];
// }

void CRC_jive::initialize_crc_table() {
  int reg;

  for (unsigned i = 0; i < 256; ++i) { // for all possible byte values
    reg = i << 8; // the first shift in the for loop will make room for the poly to be XOR'ed
    for (int j = 0; j < 8; ++j)    { // for all bits in a byte
      reg <<= 1;
      if (reg & 0x10000) {
	reg ^= crc_key;
      }
    }
    crc_table[i] = reg;
  }
}
