
#ifndef __PLTUTL__
#define __PLTUTL__
/** @brief PLOT Utility for pgplot library
 * @author M.Sekido 
 * @Date $Date: 2013-07-03 20:40:14 +0900 (Wed, 03 Jul 2013) $
 * @Version $Rev: 56 $
 * $Id: pltutl.hpp 56 2013-07-03 11:40:14Z vlbi $ 
 */

#include <string>
#include <vector>
#include <valarray>
//using namespace std;
#include 	<cstdio>
#include 	<cstdlib>
#include 	<cstring>
#include	<cmath>

extern "C" {
#include	"cpgplot.h" 
}

#include	"maxmin_v2.hpp"
#include        "cmplx.hpp"

#define  LARGER(A,B)  ((A) > (B) ? (A): (B))
#define  SMALER(A,B)  ((A) < (B) ? (A): (B))

#define  F_PHA   1
#define  F_LAG   2

/** Flags for PLT_INF ***/
#define  YMIN_FORCE   0x0001
#define  YMAX_FORCE   0x0002
#define  XMIN_FORCE   0x0004
#define  XMAX_FORCE   0x0008
#define  DEVICE_SET   0x0010
#define  PLOT_COLMN   0x0020 /* Plot in column direction (c2dplot)*/
#define  PLOT_MODE    0x0040 /* 0: line, 1: point*/
#define  XTICK_FORCE  0x0080 
#define  YTICK_FORCE  0x0100 
#define  X2MIN_FORCE  0x0200
#define  X2MAX_FORCE  0x0400
#define  X2TICK_FORCE 0x0800 
#define  PMIN_FORCE   0x1000 
#define  PMAX_FORCE   0x2000 
 /* Definition for flag2 */
#define  XOPT_FORCE   0x0001
#define  YOPT_FORCE   0x0002
#define  DRAW_MAXLINE 0x0004
#define  XYSAME_ASPCT 0x0008
#define  SIDE_PLOT    0x0010  // Plot the second line ot the max.

/** Class for PGPLOT library */
class Plt_inf {
  /** pointer to FCpmplex array for float complex value */
  const FComplex *carr;
  /** Array size for 1 and 2 dimensional plot */
  long   ndata, ndata2;
  /** Line width and Line type */
  long   lwidth, lptype;
  /** Line style parameter */ 
  int  ils; /* Line Style */
  /** Char height parameter */ 
  float cheight; /* Char height */
  /** Color index */ 
  int  icol; /* Color */
  /** Label for X/Y/Tile */ 
  char  xlabel[256],
    ylabel[256],
    title[256];
  /** flag value to control the plotting */
  unsigned long flag, flag2;
  /**  Max/Min values for Y/X and float(Z of 3D), and Phase value */
  float ymin, ymax, xmin, xmax, fmax, fmin, pmax, pmin;
  /** Tick values for second axis in 3d plot (contour,etc..)*/
  float x2min, x2max; // second axis in 3d plot (contour,etc..)  
  /** Tick values for X/Y axis */
  float xtick, ytick, x2tick;
  /** index value for MAX/MIN of Xarray,Yarray */
  int ixmax, ixmin, iymax, iymin;
  /** Array for label Option parameters for Title/X/Y axis */
  char  opt[80], xopt[20], yopt[20];
  /** Bright and Contrast parameters for color image plotting */ 
  float bright, contra;
  /** label position values */
  float lbl_pos[3][3];
  /** Sevice names for output */
  std::string device_name;
  /** Integer for palet number */
  int ipalet;
  /**flag for 
     -- 0: use cpgpanl, 
     -- 1: use cpgask(0);pgpage(); */ 
  int flag_nextpage; // 0: use cpgpanl, 1: use cpgask(0);pgpage();
  /** Function for Prameter file reading */
  void Read_Parfile(void);
 public:
  /** Default Constructor */
  Plt_inf(void);
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const std::vector <float> & xarr, const std::vector <float>& yarr);
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const std::vector <float> & xarr, const std::vector <float>& yarr,
	     const std::vector <float>& yerr );
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const std::vector <float> & xarr, const std::vector <float>& yarr,
	     const float* yerr );
  //  void Plot1(const std::vector <double> & xarr, const std::vector <double>& yarr);
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const std::vector <double> & xarr, const std::vector <double>& yarr,
	     const std::vector <double>& yerr);
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const std::vector <double> & xarr, const std::vector <double>& yarr,
	     const float* yerr=NULL);
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const long n, const float* xarr, const float* yarr,
	     const float* yerr=NULL );
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const long n, const double* xarr, const double* yarr,
	     const double* yerr);
  /** Function of plotting 1 d x/y array */ 
  void Plot1(const long n, const double* xarr, const double* yarr,
	     const float* yerr=NULL);
  //  void Oplot1(const std::vector <float> & xarr, const std::vector <float> & yarr);
  /** Function of Over plotting 1 d x/y array */ 
  void Oplot1(const std::vector <float> & xarr, const std::vector <float> & yarr,
	      const std::vector <float> & yerr );
  /** Function of Over plotting 1 d x/y array */ 
  void Oplot1(const std::vector <float> & xarr, const std::vector <float> & yarr,
	      const float* yerr=NULL );
  /** Function of Over plotting 1 d x/y array */ 
  //  void Oplot1(const std::vector <double> & xarr, const std::vector <double> & yarr);
  void Oplot1(const std::vector <double> & xarr, const std::vector <double> & yarr,
	      const std::vector <double> & yerr);
  /** Function of Over plotting 1 d x/y array */ 
  void Oplot1(const std::vector <double> & xarr, const std::vector <double> & yarr,
	      const float* yerr=NULL);
  /** Function of Over plotting 1 d x/y array */ 
  void Oplot1(const long n, const double* xarr, const double* yarr,
	      const double* yerr=NULL);
  /** Function of Over plotting 1 d x/y array */ 
  void Oplot1(const long n, const double* xarr, const double* yarr,
	      const float* yerr);
  /** Function of Over plotting 1 d x/y array */ 
  void Oplot1(const long n, const float* xarr, const float* yarr,
	      const float* yerr=NULL);
  /** Setting and Opening the device file */ 
  void Dev_name(std::string dev="");
  /** Add options for X/Y label */
  void Add_opt(const char *xstr, const char *ystr){ 
    strcpy(xopt,xstr);
    strcpy(yopt,ystr);
  }
  /** Resetting options for X/Y label*/ 
  void Reset_opt(void){ xopt[0] ='\0'; yopt[0] ='\0'; }
  /** Computing and returning tick value */
  float Tickcal(float tick);
  /** Setup Character height */
  void Set_cheight(const float ch){cheight =ch;}
  /** Setup Symbol parameter for plot */
  void Set_symbol(const int sym){flag |= PLOT_MODE; lptype = sym; }
  /** Setup Lyne style parameter for plot */
  void Set_line(int l){flag &= ~PLOT_MODE; ils=l;}
  /** Setup Line width parameter for plot */
  void Set_lw(int l){lwidth=l;}
  /** Setup Bright and Contrast parameter for plot */
  void Set_Bright_cntr(float br, float cntr){bright=br; contra = cntr;}
  /** Setup Color parameter for plot */
  void Set_color(int ic){icol=ic;}
  /** Setup Window aspect  parameter for plot */
  void Set_wnad(int ic){
    if( ic) flag2 |= XYSAME_ASPCT; 
    else flag2 &= ~XYSAME_ASPCT;
  }
  /** Setup Flag for plot */
  void Set_Flag2(const unsigned long i){ flag2 |= i;}
  /** Setup lablel values for X/Y/Title  */
  void Set_label(const char *xlabel, const char *ylabel = "", 
		 const char * title =""){ 
    strcpy(this->xlabel, xlabel);
    strcpy(this->ylabel, ylabel);
    strcpy(this->title, title);
  }
  /** Set to plot column data in 2d plotting    */
  void Set_pclmn(int s){
    if(s)  flag |= PLOT_COLMN;
    else   flag &= ~PLOT_COLMN;
  }
  /** Setup palet index     */
  void Set_palette(int ip){ipalet = ((ip-1) % 5) +1 ; Palette();}
  /** Setup X-min     */
  void Set_xmin(float xmn){ xmin = xmn; flag |= XMIN_FORCE;}
  /** Setup X-max     */
  void Set_xmax(float xmx){ xmax = xmx; flag |= XMAX_FORCE;}
  /** Setup Y-min     */
  void Set_ymin(float ymn){ ymin = ymn; flag |= YMIN_FORCE;}
  /** Setup Y-max     */
  void Set_ymax(float ymx){ ymax = ymx; flag |= YMAX_FORCE;}
  /** Setup Phasse-min     */
  void Set_pmin(float ymn){ pmin = ymn; flag |= PMIN_FORCE;}
  /** Setup Phase-max     */
  void Set_pmax(float ymx){ pmax = ymx; flag |= PMAX_FORCE;}
  /** Setup X-tick value   */
  void Set_xtick(float xtk){ xtick = xtk; flag |= XTICK_FORCE;}
  /** Setup Y-tick value   */
  void Set_ytick(float ytk){ ytick = ytk; flag |= YTICK_FORCE;}
  /** UNSetup X-min     */
  void Unset_xmin(void){ flag &= ~XMIN_FORCE;}
  /** UNSetup X-max     */
  void Unset_xmax(void){ flag &= ~XMAX_FORCE;}
  /** UNSetup Y-min     */
  void Unset_ymin(void){ flag &= ~YMIN_FORCE;}
  /** UNSetup Y-max     */
  void Unset_ymax(void){ flag &= ~YMAX_FORCE;}
  /** UNSetup X-tick value   */
  void Unset_xtick(void){ flag &= ~XTICK_FORCE;}
  /** UNSetup Y-tick value   */
  void Unset_ytick(void){ flag &= ~YTICK_FORCE;}
  unsigned char Chk_xminforce(){return(flag&XMIN_FORCE);}
  unsigned char Chk_xmaxforce(){return((flag&XMAX_FORCE) >> 1);}
  unsigned char Chk_yminforce(){return((flag&YMIN_FORCE) >> 2);}
  unsigned char Chk_ymaxforce(){return((flag&YMAX_FORCE) >> 3);}
  float Ret_xmin(){return(xmin);}
  float Ret_xmax(){return(xmax);}
  float Ret_ymin(){return(ymin);}
  float Ret_ymax(){return(ymax);}
  float Ret_fmax(){return(fmax);}
  float Ret_fmin(){return(fmin);}
  void Set_lbl_pos(int i, float a, float b = 0.5, float c=0.5){
    // i= 0:"T", 1:"B", 2:"L" 
    lbl_pos[i][0] =a;lbl_pos[i][1] =b;lbl_pos[i][2] =c; 
  }
  int Chk_pclmn(void){return((flag & PLOT_COLMN)>>5);}
  /** Setup palette values */
  void Palette(void);
  /** Plot 2-D values */
  void Plot2d(const float *xarr, const float *yarr, const float *farr, int dnx, int dny );
  /** Plot 2-D Complex values */
  int C2dplot(const float*, const float*, const FComplex*, long, long);
  // void Plot2d2(float *xarr, float *yarr, float *farr, int dnx, int dny );
  void Dev_set(void){flag |= DEVICE_SET;} /*use output setup file*/
  int Dev_chk(void){return((flag & DEVICE_SET) >>1);}
  /** Returning index of max in x-data */
  long Ixmax(void){return(ixmax);}
  /** Returning index of max in y-data */
  long Iymax(void){return(iymax);}
  /** Returning character pointer for title */
  char* Title(void){ return(title);}
  void Set_nextpage(int i=0){flag_nextpage = i;}
  void Set_cpval(const float x); // Set color pallette value 
};

inline void Plt_inf::Set_cpval(const float x){
  float crange =0;
  float x1 =x;
  if(x<0) x1=0;
  if(x>1) x1=1;
  switch (ipalet){
  case 1: // Gray scale
    crange =1;    break;
  case 2:
    crange = 9;    break;
  case 3:
    crange = 5;    break;
  case 4:
    crange = 10; break;
  case 5:
    crange = 20; break;
  }
  int icol = (int)rint(x1*crange); 
  icol = icol == 0 ? 1 : icol;
  cpgsci(icol);     
}

inline Plt_inf::Plt_inf(void):
  ndata(0),ndata2(0),lwidth(2),lptype(1),ils(1),cheight(2.0),icol(1),
  //  xlabel(""),ylabel(""),title(""), 
  flag(0),flag2(0), 
  ymin(0),ymax(0),xmin(0),xmax(0),fmax(0),fmin(0),pmax(0),pmin(0),
  x2min(0), x2max(0), xtick(0),ytick(0),x2tick(0),
  ixmax(0),ixmin(0),iymax(0),iymin(0),
  //  opt(""),xopt(""),yopt(""),
  bright(0.75),contra(3.5),
  ipalet(2), flag_nextpage(0){ 
//     flag=0;  flag2=0;
//  opt[0] = xopt[0] = yopt[0] = '\0';
//     lptype = 1; /* Default symbol */
//     lwidth=2;
//     ils =1; icol =1;
//     cheight=2.0;
    lbl_pos[0][0] = 1.;  // Top
    lbl_pos[0][1] = 0.5;
    lbl_pos[0][2] = 0.5; 
    lbl_pos[1][0] = 2.5; // Bottom 
    lbl_pos[1][1] = 0.5;
    lbl_pos[1][2] = 0.5; 
    lbl_pos[2][0] = 2.; // Left 
    lbl_pos[2][1] = 0.5;
    lbl_pos[2][2] = 0.5; 
//     ipalet =2; // Rainbow 
//     contra =3.5; // 0-10.
//     bright = 0.75; // 0-1.
//     flag_nextpage=0;
}

inline void Plt_inf::Plot1(const std::vector <float> & xarr, 
			   const std::vector <float>& yarr)
{ Plot1(xarr.size(), &xarr[0], &yarr[0]); }

inline  void Plt_inf::Plot1(const std::vector <float> & xarr, 
			    const std::vector <float>& yarr,
			    const std::vector <float>& yerr )
{  Plot1(xarr.size(), &xarr[0], &yarr[0], &yerr[0]); }

inline void Plt_inf::Plot1(const std::vector <float> & xarr, 
			   const std::vector <float>& yarr,
			   const float* yerr )
{ Plot1(xarr.size(), &xarr[0], &yarr[0], yerr); }

inline void Plt_inf::Plot1(const std::vector <double> & xarr, 
			  const std::vector <double>& yarr,
			  const std::vector <double>& yerr)
{  Plot1(xarr.size(), &xarr[0], &yarr[0], &yerr[0]); }

inline void Plt_inf::Plot1(const std::vector <double>& xarr, 
			   const std::vector <double>& yarr,
			   const float* yerr){  
  long ndat = xarr.size();
  std::vector <float> x;
  x.resize(ndat);
  std::vector <float> y;
  y.resize(ndat);
  
  for(long i=0; i<ndat; i++){
    x[i] = xarr[i];
    y[i] = yarr[i];
  }
  if(yerr == NULL){
    Plot1(xarr.size(), &x[0], &y[0]);     
  }else{
    Plot1(xarr.size(), &x[0], &y[0], yerr); 
  }

}

inline void Plt_inf::Oplot1(const std::vector <double> & xarr, 
			  const std::vector <double>& yarr,
			  const float* yerr){ 

  long ndat = xarr.size();
  std::vector <float> x(ndat);
  std::vector <float> y(ndat);
  for(long i=0; i<ndat; i++){
    x[i] = xarr[i];
    y[i] = yarr[i];
  }

  if( yerr == NULL){
    Oplot1(xarr.size(), &x[0], &y[0]); 
  }else{ 
    Oplot1(xarr.size(), &x[0], &y[0], yerr); 
  }
}

inline void Plt_inf::Oplot1(const long ndat, const double *xarr, 
			    const double *yarr, const double *yerr){
  
  if( yerr == NULL){
    std::vector <float> x(ndat);
    std::vector <float> y(ndat);
    for(long i=0; i<ndat; i++){
      x[i] = xarr[i];
      y[i] = yarr[i];
    }

    Oplot1(ndat,&x[0], &y[0]);
  }else{
    std::vector <float> x(ndat);
    std::vector <float> y(ndat);
    std::vector <float> ye(ndat);
    for(long i=0; i<ndat; i++){
      x[i] = xarr[i];
      y[i] = yarr[i];
      ye[i] = yerr[i];
    }
   Oplot1(ndat,&x[0], &y[0], &ye[0]);
  }
  
}
/*
inline void Plt_inf::Oplot1(const std::vector <float> & xarr, 
			    const std::vector <float> & yarr)
{  Oplot1(xarr.size(), &xarr[0], &yarr[0]);  }
*/

inline void Plt_inf::Oplot1(const std::vector <float> & xarr, 
			    const std::vector <float> & yarr,
			    const std::vector <float> & yerr )
{  Oplot1(xarr.size(), &xarr[0], &yarr[0], &yerr[0]);  }

inline void Plt_inf::Oplot1(const std::vector <float> & xarr, 
			    const std::vector <float> & yarr,
			    const float* yerr )
{  
  if(yerr ==NULL)    
    Oplot1(xarr.size(), &xarr[0], &yarr[0]); 
  else
    Oplot1(xarr.size(), &xarr[0], &yarr[0], yerr); 
} 


inline void Plt_inf::Oplot1(const std::vector <double> & xarr, 
			    const std::vector <double> & yarr,
			    const std::vector <double> & yerr)
{ Oplot1(xarr.size(), &xarr[0], &yarr[0], &yerr[0]); }

/** Color Line 
 * iicol: Initial color index 
 * ifcol: Final color index
 * carr[]: Character for sample, or sequential number in case of 0.
 */
void Color_Line_Sample(int iicol,  int ifcol,
		       const float xoff=0, const float yoff=0,
		       char  *arr[]=0);

/*** Prot types of pltutl.c ********/
/*
int f2dplot(void);
void cont1(float farr[], int nx, int ny, float xmin, float ymin,  
	   float xstp, float ystp, char xtitle[], char xlabel[], 
	   char ytitle[], char ylabel[]);
void palette(int type, float bright, float contra);
void colorcnt(float farr[], long nx, long ny, 
	      char xlabel[], char ylabel[], char title[]);

void plotc1ap(PLT_INF plt_inf);
void plotc1ri(complex carr[], int ndata, char title[] );
void plot1(float farr[], int ndata, char titile[] );


float	tickcal(float tick);
void 	plotlpt(float ypt[], int npt, float xl[], float yl[], int nl );

*/

#endif
