#!/usr/bin/perl -w

$port = 9902;

sub usage(){
    print "Usage: $0 <-a <host>[:port]>  <-m Msps:bit:ch:firstch:ngrp> \n";
    print "\t  <-o <dir>> [-c <config>]  [-t <sec>]\n";
    print "  Test writing data with vsi_localwrite_srv via remote control.\n";
    print "    -m: mode setting. ex) 16:1:4:0:4  is 16Msps/1bit/4ch for 4 group starting from 0 ch\n";
    print "    -t: duration of test writing. Default=10sec.\n";
    print "    -a: Default port Number of $port.\n";
    print "    -c: Reading Configfile, ignoring other arguments(-o/a/m).\n";

    exit(1);
}

# use Exporter;
use IO::Socket;                   # IO::Socket ⥸塼Ȥ
use lib "$ENV{HOME}/prog/perl/module";
use Time_lib;
use Conf_hash;
{

    my ($host,$mode,$outdir,$dur);
    $dur=10;
    my $confile;
    
    while(@ARGV){
	my $opt =shift;
	if($opt =~/-a/){
	    $host = shift;
	}elsif($opt =~/-m/){
	    $mode = shift;
	}elsif($opt =~/-o/){
	    $outdir = shift;
	}elsif($opt =~/-t/){
	    $dur = shift;
	}elsif($opt =~/-c/){
	    $confile = shift;
	}
    }
    if(defined($confile)){
	my $conf_dat = new Conf_hash;
	if ($conf_dat->Read($confile) < 1){
	    print "Can't open a config file: $!\n";
	    usage();
	}
	$outdir  = $conf_dat->Value("Outdir");
	$host  = $conf_dat->Value("RecordHost");
	$port  = $conf_dat->Value("RecPortNum");
	my $mode1  = $conf_dat->Value("RecMode");
	my @trm = split(/:/,$mode1);
	my $nch = $trm[2];
	my $firstch = $conf_dat->Value("Firstch");
	my $nch_total  = $conf_dat->Value("NchTotal");
	print "total/nch= $nch_total/$nch \n";
	my $ngrp = $nch_total/$nch;
	$mode=sprintf("%s:%d:%d",$mode1,$firstch,$ngrp);
    }
    

    &usage() unless(defined($host) && defined($outdir)
		   && defined($mode) );
    if($host =~/:/){
	my @terms = split(/:/,$host);
	$host = $terms[0];
	$port = $terms[1];
    }
    print "Host/port= $host : $port \n";
    my $remote = IO::Socket::INET->new(Proto    => "tcp",
				       PeerAddr => $host,
				       PeerPort => $port );
    if( ! $remote){	die "Could not connect to $host:$port \n";    }
    select($remote); $|=1; # stop buffering 
    select(STDOUT); $|=1; # stop buffering 

    my $ret = &Set_param($remote, $outdir, $mode);
    if($ret < 0){
	print STDERR "E Failed Set_param().\n";
	exit(1);
    }
    
    my $cmd = "in2net=on;";
    printf "$cmd\nRet: %s \n", &Cmd($remote,$cmd);
    print "Sleeping $dur sec....\n";
    sleep($dur);
    $cmd = "in2net=off;";
    printf "$cmd\nRet: %s \n", &Cmd($remote,$cmd);
    print "Finished.\n";
}

sub Cmd($$){
    my ($remote,$cmd) = @_;
    print $remote $cmd;
    my $ret=<$remote>;
#    while( <$remote>){	$ret = "$ret,  $_";    }
    
    return $ret;
}


sub Set_param($$$){
    my $remote = shift;
    my $outdir = shift;
    my $mode = shift;
    my @param = split(/:/,$mode);
    my ($Mbps,$bit,$ch,$first,$ngrp) = @param;
    unless(defined($ngrp)){
	print STDERR "E mode parameter is not 5 terms as expected.";
	return -1;
    }
    
    my $rate = $Mbps*$bit*$ch;
    my $cmd = sprintf("play_rate=data:%d;",$rate);
    my $ret = &Cmd($remote, $cmd);
    print "$cmd\nRet: $ret \n"; 
    $cmd = sprintf("data_out=multi:%d;",$ngrp);
    $ret = &Cmd($remote, $cmd);
    print "$cmd\nRet: $ret \n";

    my $tm = new Time_lib;
    my $ifirst = $first;
    for(my $i=0; $i<$ngrp; $i++){
	$tm->Set_time_now();
	my $fname = sprintf("%s/test%s-%d.dat",$outdir,$tm->Str_fmt1(33),$i);
	$cmd = sprintf("data_out=FILE%d:%s;",$i,$fname);
	printf "$cmd\nRet: %s \n", &Cmd($remote, $cmd);

	$cmd = sprintf("modegrp=%d:%d:%d:%d:%d;",$i,$Mbps,$bit,$ch,$ifirst);
	printf "$cmd\nRet: %s \n", &Cmd($remote, $cmd);
	$ifirst += $ch;
    }
    
    return 0;
}


